"""
This module contains functionality to load MUI -> h5 converted files in a standardized way. The data loader
is based on the Dataloader template
"""

import logging
import math
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

import h5py
import numpy as np
import numpy.typing as npt
import scipy.constants

from radardef.components import DataLoader
from radardef.radar_stations.mu.validators import H5
from radardef.types import BoundParams, ExpParams, Metadata, Pointing, TargetFormat


class H5Loader(DataLoader):
    """Simplifies the way to load .h5 files converted from MUI"""

    __logger = logging.getLogger(__name__)

    @property
    def meta(self) -> Metadata:
        """Metadata, containing experiment data and bounds data"""
        return self.__meta

    @property
    def pointing(self) -> Pointing:
        """Pointing data, data describing the radar pointing direction in spherical coordinates"""
        return Pointing(data=[(0, {"azimuth": 0.0, "elevation": 90.0})], sample_rate=1)

    @property
    def channels(self) -> list[int] | list[str]:
        """All available channels"""
        return self.meta.experiment.rx_channels

    def __init__(self) -> None:
        super().__init__(TargetFormat.H5, H5())
        self.__meta = Metadata(ExpParams(), BoundParams())
        self.__sample_bounds: dict[str | int, tuple[int, int]] = {}
        self.__t_ipp_usec = 3120
        self.__t_rx_start_usec = 486
        self.__t_samp_usec = 6
        self.__sample_rate = 1 / (self.__t_samp_usec * 1e-6)
        self.__radar_frequency = 46.5  # TODO: h5file.attrs["tx_frequency"] returns a 0 array
        self._ipp_samps = int(self.__t_ipp_usec / self.__t_samp_usec)
        self.__code = np.kron(
            np.array(
                [1, 1, 1, 1, 1, -1, -1, 1, 1, -1, 1, -1, 1],
                dtype=np.float64,
            ),
            np.ones(2),
        ).astype(np.float64)
        self.__samples_per_file = 266240

    def load(self, path: Path | str) -> None:
        """Loads a path to the dataloader, extracting metadata and other important specifications

        Args:
            path: path to data file

        """

        self.__path = Path(path).resolve()
        if self.__path.is_dir():
            files = self._get_all_files_from_dir(self.__path)
            self.__meta, self.__sample_bounds = self._extract_meta_from_list(files)
        else:
            self.__meta, self.__sample_bounds = self._extract_meta(self.__path)

    def bounds(self, channel: str | int) -> tuple[int, int]:
        """Sample bounds of the specific channel

        Args:
            channel: channel
        Returns:
            tuple of start and end sample
        Raises:
            Exception: channel is missing
        """

        if not isinstance(channel, int):
            channel = int(channel)

        if self._is_channel_present(channel):
            return self.__sample_bounds[channel]
        else:
            raise Exception(f"channel {channel} is missing in {dir}")

    def read(
        self,
        channel: str | int,
        start_sample: Optional[int] = None,
        vector_length: Optional[int] = None,
    ) -> npt.NDArray[np.complex128]:
        """
        Read data from loaded file

        Args:
            channel (optional): Channel to read data from
            start_sample (optional): Start of range to read, if empty all data will be read
            vector_length (optional): Number of samples (counting from start sample) to read,
                                    if empty all data will be read

        Returns:
            Complex data of length vector_length from give channel

        Raises:
            Exception: channel is missing

        NOTE:
            Ipp structure:

            ```
            0                   26       81                 166                520
            |    26 samples     |         |    85 samples    |
            |tx_start-----tx_end|_________|rx_start----rx_end|__________________|
            |---------------------------------ipp-------------------------------|
            ```

            included in raw data:
            ```
                |rx_start-----rx_end|
            ```
            padded data:
            ```
                | 81 zeros | 85 samples rx data | 354 zeros |
            ```
        """

        if not isinstance(channel, int):
            channel = int(channel)

        if not self._is_channel_present(channel):
            raise Exception(f"channel {channel} is missing in {dir}")

        if self.__path.is_dir():
            files = self._get_all_files_from_dir(self.__path)

            if start_sample is not None:
                start_file = math.floor(start_sample / self.__samples_per_file)
                index = start_sample % self.__samples_per_file
            else:
                start_file = 0
                index = 0
                start_sample = 0

            if vector_length is not None:
                num_files = math.ceil(
                    ((start_sample % self.__samples_per_file) + vector_length) / self.__samples_per_file
                )
                samples = vector_length
            else:
                num_files = len(files)
                samples = self.bounds(channel)[1]

            files = files[start_file : start_file + num_files]
            padded_data = np.empty((0,), dtype=np.complex128)
            for i, file in enumerate(files):
                h5file = self._open_h5_file(file)
                data = h5file["data"][channel - 1]
                h5file.close()
                if i == 0:
                    padded_data = self._flatten_and_zero_pad(data)
                else:
                    padded_data = np.concatenate((padded_data, self._flatten_and_zero_pad(data)), axis=0)

            return padded_data[index : index + samples]
        else:
            h5file = self._open_h5_file(self.__path)
            data = h5file["data"][channel - 1]
            h5file.close()

            # flatten and fill with zeroes
            padded_data = self._flatten_and_zero_pad(data)

            if start_sample is None and vector_length is None:
                return padded_data
            elif start_sample is not None and vector_length is not None:
                return padded_data[start_sample : start_sample + vector_length]
            elif start_sample is None and vector_length is not None:
                return padded_data[0:vector_length]
            else:
                return padded_data[start_sample:]

    def _flatten_and_zero_pad(self, data: npt.NDArray[np.complex128]) -> npt.NDArray[np.complex128]:
        """
        Zero pads the data to make it match the true signal
        ```
        | 81 zeros | 85 samples rx data | 354 zeros |
        ```
        """

        pulses = len(data)
        rx_start_samp = int(self.meta.experiment.t_rx_start_usec / self.meta.experiment.t_samp_usec)

        padded_data = np.zeros((self._ipp_samps * pulses), dtype=np.complex128)

        for ipp_n, rx_batch in enumerate(data):
            offset = int(ipp_n * self._ipp_samps) + rx_start_samp
            padded_data[offset : offset + len(rx_batch)] = rx_batch

        return padded_data

    def _open_h5_file(self, path: Path) -> h5py.File:
        """Open h5 file and return reader"""

        try:
            h5file = h5py.File(str(path), "r")
        except FileNotFoundError:
            self.__logger.exception(f"Could not open file: {path}. File does not exist.")
            raise
        except OSError:
            self.__logger.exception(f"File {path} was not a h5 file, and was probably in binary format.")
            raise
        except UnicodeDecodeError:
            self.__logger.exception(f"File {path} was not a h5 file.")
            raise

        return h5file

    def _get_all_files_from_dir(self, path: Path) -> list[Path]:
        """Get all files available in dir"""
        paths: list[Path] = []
        if not path.is_dir():
            return paths

        paths = [f for f in path.iterdir() if f.is_file() and self.validate(f)]
        paths.sort()

        if len(paths) == 0:
            raise Exception(f"No valid h5 files at: {path}")
        return paths

    def _extract_meta(self, path: Path) -> tuple[Metadata, dict[str | int, tuple[int, int]]]:
        """Get metadata from h5 file"""

        h5file = self._open_h5_file(path)

        experiment = ExpParams(
            name=h5file.attrs["filename"],
            radar_frequency=self.__radar_frequency,
            t_ipp_usec=self.__t_ipp_usec,
            sample_rate=self.__sample_rate,
            ipp_samps=self._ipp_samps,
            t_samp_usec=self.__t_samp_usec,
            t_rx_start_usec=self.__t_rx_start_usec,
            t_rx_end_usec=self.__t_rx_start_usec + 6 * 85,
            t_tx_start_usec=0,
            t_tx_end_usec=26 * self.__t_samp_usec,  # code length * t_samp
            rx_channels=h5file["rx_channels"][()],
            pulse=2,
            code=self.__code,
            wavelength=scipy.constants.c / (self.__radar_frequency * 1e6),
        )

        start_time = datetime.strptime(str(h5file.attrs["record_start_time"])[0:26], "%Y-%m-%dT%H:%M:%S.%f")
        end_time = datetime.strptime(str(h5file.attrs["record_end_time"])[0:26], "%Y-%m-%dT%H:%M:%S.%f")
        # time can be validated by end_time = start_time.timestamp() + (n_ipp * t_ipp_usec) * 1e-6, n_ipp = 512
        # if this is not matching maybe the measurement stopped  early
        bounds = BoundParams(
            ts_start_usec=start_time.timestamp(),
            ts_end_usec=end_time.timestamp(),
        )

        sample_bounds: dict[str | int, tuple[int, int]] = {}
        for channel, data in enumerate(h5file["data"]):
            pulses = len(data)
            ipp_length = int(experiment.t_ipp_usec / experiment.t_samp_usec)
            sample_bounds[channel + 1] = (0, pulses * ipp_length)
        h5file.close()

        return Metadata(experiment, bounds), sample_bounds

    def _extract_meta_from_list(self, paths: list[Path]) -> tuple[Metadata, dict[str | int, tuple[int, int]]]:
        """Concatenate metadata from several h5 file"""

        sample_bounds: dict[str | int, tuple[int, int]] = {}
        experiment = ExpParams()

        # sort accoring to timestamp
        paths.sort()
        for i, f in enumerate(paths):
            h5file = self._open_h5_file(f)

            if i == 0:
                experiment = ExpParams(
                    name=h5file.attrs["filename"],
                    radar_frequency=self.__radar_frequency,
                    t_ipp_usec=self.__t_ipp_usec,
                    sample_rate=self.__sample_rate,
                    ipp_samps=self._ipp_samps,
                    t_samp_usec=self.__t_samp_usec,
                    t_rx_start_usec=self.__t_rx_start_usec,
                    t_rx_end_usec=self.__t_rx_start_usec + 6 * 85,
                    t_tx_start_usec=0,
                    t_tx_end_usec=27 * self.__t_samp_usec,  # (code length + 1) * t_samp,
                    rx_channels=h5file["rx_channels"][()],
                    pulse=2,
                    code=np.kron(
                        np.array(
                            [1, 1, 1, 1, 1, -1, -1, 1, 1, -1, 1, -1, 1],
                            dtype=np.float64,
                        ),
                        np.ones(2, dtype=np.float64),
                    ).astype(np.float64),
                    wavelength=scipy.constants.c / (self.__radar_frequency * 1e6),
                )

                start_time = datetime.strptime(
                    str(h5file.attrs["record_start_time"])[0:26], "%Y-%m-%dT%H:%M:%S.%f"
                )

            elif i == (len(paths) - 1):
                end_time = datetime.strptime(
                    str(h5file.attrs["record_end_time"])[0:26], "%Y-%m-%dT%H:%M:%S.%f"
                )

            ipp_length = int(experiment.t_ipp_usec / experiment.t_samp_usec)
            for i, data in enumerate(h5file["data"]):
                channel = i + 1
                if channel in experiment.rx_channels:
                    min_max = (0, len(data) * ipp_length)
                    if channel not in sample_bounds:
                        sample_bounds[channel] = min_max
                    else:
                        sample_bounds[channel] = tuple(np.add(sample_bounds[channel], min_max))

            h5file.close()

        bounds = BoundParams(
            ts_start_usec=start_time.timestamp(),
            ts_end_usec=end_time.timestamp(),
        )

        return Metadata(experiment, bounds), sample_bounds

    def _is_channel_present(self, chnl: str | int) -> bool:
        """is channel present in the data"""
        return chnl in self.meta.experiment.rx_channels



"""Eiscat Svalbard Radar (ESR) object"""

from radardef.components.radar_station_template import RadarStation
from radardef.radar_stations.eiscat.beams.esr import (
    esr_32m_cassegrain_beam,
    esr_42m_cassegrain_beam,
)
from radardef.types import DishDiameter, StationID

from .converters import EiscatMatbzToDrf
from .data_loaders import DrfLoader
from .validators import EiscatMatlab


class ESR(RadarStation):
    """
    Eiscat Svalbard Radar (ESR) definition

    Args:
        dish_diameter: Dish diameter (32m, 42m)

    """

    def __init__(
        self,
        dish_diameter: DishDiameter = DishDiameter.ESR_32M,
    ) -> None:

        if dish_diameter == DishDiameter.ESR_32M:
            beam, params = esr_32m_cassegrain_beam()
            station_id = StationID.ESR_32M
            lon = 16.0758715
            min_elevation = 15
        elif dish_diameter == DishDiameter.ESR_42M:
            beam, params = esr_42m_cassegrain_beam()
            station_id = StationID.ESR_42M
            lon = 16.081483
            min_elevation = None

        super().__init__(
            station_id,
            transmitter=True,
            receiver=True,
            lat=78.153145,
            lon=lon,
            alt=185,
            beam=beam,
            beam_parameters=params,
            min_elevation=min_elevation,
            noise_temperature=70,
            power=1e6,
            frequency=500e6,
            converters=[EiscatMatbzToDrf()],
            validator=EiscatMatlab(),
            data_loaders=[DrfLoader()],
        )



"""Eiscat mat.bz2 utility functions."""

import bz2
import datetime as dt
import re
from pathlib import Path
from typing import Any, Callable, Generator, Mapping, Optional

import numpy as np
import scipy.io as sio
from numpy.typing import NDArray

from radardef.radar_stations.eiscat.utils import load_expconfig
from radardef.radar_stations.eiscat.utils.drf_utils import str_from_ts

PARBL_ELEVATION: int = 8
PARBL_AZIMUTH: int = 9
PARBL_END_TIME: int = 10
PARBL_SEQUENCE: int = 11
PARBL_START_TIME: int = 42  # upar[1]
PARBL_RADAR_FREQUENCY: int = 54  # upar[13]


def index_of(alist: list[Path], match_func: Callable[[Path], bool]) -> int:
    """
    Returns index of first object in list which matches given match func
    """
    matches = [match_func(e) for e in alist]
    return matches.index(True)


def to_i2x16(zz: NDArray[np.complex128]) -> NDArray[np.int16]:
    """
    convert from (N,) complex128 to (N,2) int16
    """
    zz2x16 = np.empty((len(zz), 2), dtype=np.int16)
    zz2x16[:, 0] = zz.real.astype(np.int16)
    zz2x16[:, 1] = zz.imag.astype(np.int16)
    return zz2x16


def loadmat(path: str) -> Mapping[str, NDArray[np.complex128]]:
    """
    Load matlab file.
    Transparently unzips files on the fly (not in the filesystem)
    """
    if path.endswith(".mat.bz2"):
        return sio.loadmat(bz2.open(path, "rb"))
    else:
        return sio.loadmat(path)


def expinfo_split(xpinf: str) -> tuple[str, ...]:
    """
    Move from hard coded constants to loading config based on exp name/version
    'kst0 leo_bpark_2.1u_NO' -> ('kst0', 'leo_bpark', '2.1u', 'NO')
    """
    try:
        # host, name, versi, owner = \
        match = re.match(r"(\w+) +(\w+)_(\d+(?:\.\d+)?[a-z]*)_(\w+)", xpinf)
        if match is not None:
            return match.groups()
        else:
            return "", "", "", ""
    except Exception as e:
        raise ValueError(f"d_ExpInfo: {xpinf} not understood: {e}")


def parse_foldername(product_folder: str | Path) -> tuple[str, str, str]:
    """
    Parses foldername of eiscat raw data product to produce (expname, expvers, chnl)
        e.g. foldername "leo_pwait_2.3r_3P@sod"
        e.g. result ('leo_pwait', '2.3r', 'sod')

    NOTE - this is similar to expinfo split - which produces similar output
    from parsing eiscat metadata. This function additionally
    gets the chnl name. This function would be obsolete, if correct chnl
    name was parsed from metadata instead. At this time, we are useing the
    folder as a quick solution.
    """
    product_folder = Path(product_folder)
    tokens = product_folder.name.split("_")
    expname = "_".join(tokens[:2])
    expvers = tokens[2]
    chnl = tokens[3].split("@")[1]
    return expname, expvers, chnl


def beginning_of_year(_dt: dt.datetime) -> dt.datetime:
    """Given a datetime, returns a new datetime representing the beginning of the year"""
    return _dt.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)


def get_seconds_since_year_start(_dt: dt.datetime) -> int:
    """Given a datetime, returns the number of seconds into the year"""
    _dt_year_start = beginning_of_year(_dt)
    return int((_dt - _dt_year_start).total_seconds())


def get_filedate_from_sample_file_end(sample_file_end: int, sample_rate: float) -> int:
    """Seconds into the year based on the sample file end and sample rate"""
    # filename date is calculated from file_end (first sample of next file)
    ts = sample_file_end / sample_rate
    _dt = dt.datetime.fromtimestamp(ts, tz=dt.timezone.utc)
    return get_seconds_since_year_start(_dt)


def eiscat_file_idx_by_datetime(files: list[Path], _dt: dt.datetime) -> int:
    """
    Return index of eiscat file covering ts
    """

    # check that _dt is same year as files
    year = int(files[0].parent.name[:4])
    if year != _dt.year:
        raise Exception(f"mismatch year: dt year {_dt.year}, files year {year}")

    # find index of first file which offset > offset of _dt
    offset = get_seconds_since_year_start(_dt)

    def match_start(file: Path) -> bool:
        """Gets datetime from path name and determines if it is larger than _dt"""
        _offset = int(file.name.split(".")[0])
        return _offset > offset

    return index_of(files, match_start)


def eiscat_files(
    productpath: Path,
    start: Optional[str | dt.datetime | int] = None,
    end: Optional[str | dt.datetime | int] = None,
    count: Optional[int] = None,
) -> list[Path]:
    """
    Sort eiscat files

    Args:
        productpath: Path to directory containing eiscat files
        start (optional):  int: file index, datetime: start time (utc), str: str date, year offset in seconds
        end (optional): int: file index, datetime: end time (utc), str: str date, year offset in seconds
        count (optional): Number of files to return from start, overrides end

    Returns:
        Sorted list of eiscat files
    """

    def ok(f: Path) -> bool:
        """Path is a eiscat .mat.bz2 file"""
        return f.is_file() and f.name.endswith(".mat.bz2")

    if productpath.is_dir():
        files = [f for f in productpath.iterdir() if ok(f)]
    else:
        files = [productpath]
    # sort by path (ascending in time)
    files.sort()

    start_idx = 0
    end_idx = len(files)

    # start
    if start is None:
        start_idx = 0
    elif isinstance(start, str):
        # start is str date (year offset in seconds) from filename
        def match(f: Path) -> bool:
            """match str date with path name, true if equal"""
            return start == f.name.split(".")[0]

        start_idx = index_of(files, match)
    elif isinstance(start, dt.datetime):
        start_idx = eiscat_file_idx_by_datetime(files, start)
    elif isinstance(start, int):
        # start is index
        start_idx = start
    else:
        raise Exception("illegal start", start)

    # end
    if end is not None:
        if isinstance(end, str):
            # end is str date from filename
            def match(f: Path) -> bool:
                """match str date with path name, true if equal"""
                return end == f.name.split(".")[0]

            end_idx = index_of(files, match) + 1
        elif isinstance(end, dt.datetime):
            end_idx = eiscat_file_idx_by_datetime(files, end)
        elif end is None:
            end_idx = len(files)
        elif isinstance(end, int):
            # end is index
            end_idx = end
        else:
            raise Exception("illegal end", end)

    # count has precedence over end
    if count is not None:
        end_idx = min(start_idx + count, len(files))

    return files[start_idx:end_idx]


def eiscat_load_file(
    filepath: Path,
) -> tuple[dict[str, dict[str, Any]], NDArray[np.int16], dict[str, float], list[str]]:
    """
    Returns a dictionary with meta information for given eiscat file
    """
    mat = loadmat(str(filepath))

    host, expname, expvers, owner = expinfo_split(str(mat["d_ExpInfo"][0]))
    cfg = load_expconfig(expname)
    cfv = cfg[expvers]

    sample_rate_raw = cfv.get("sample_rate")
    if sample_rate_raw is None:
        raise Exception(f"Not possible to read sample rate")
    sample_rate = float(sample_rate_raw)

    file_secs_raw = cfv.get("file_secs")
    if file_secs_raw is None:
        raise Exception(f"Not possible to read file secs")
    file_secs = float(file_secs_raw)

    samples_per_file = int(file_secs * sample_rate)
    chnl = parse_foldername(filepath.parent.parent)[2]

    # global start time for sampling (repeated for all files)
    ts_origin_sec = float(mat["d_parbl"][0][PARBL_START_TIME])
    # sample index corresponding to global start time
    idx_origin = int(np.floor(ts_origin_sec * sample_rate))
    # end time of file
    ts_endfile_sec = float(mat["d_parbl"][0][PARBL_END_TIME])

    # NOTE: ts_endfile_sec can not be trusted to be precisely
    # consistent with ts_origin_sec - as a source for sample number.
    # we use the approach of counting samples from ts_orgin_sec,
    # relying on a fixed sample count per file

    # NOTE: file_idx is a logical sequence number for files, starting from
    # file_idx 0 at ts_origin_sec. This does not correspond to the order of files processed.
    # For instance, the first file might have file_idx 6, indicating that the
    # recording only started some time after ts_origin_sec
    # also the inprecision in ts_endfile_sec goes away in division
    file_idx = round((ts_endfile_sec - ts_origin_sec) / file_secs) - 1
    # sample index for start of first file
    samples_per_file = int(file_secs * sample_rate)
    # index of filestart
    idx_startfile = idx_origin + file_idx * samples_per_file

    # check that samples is correct
    data = to_i2x16(mat["d_raw"][:, 0])
    n_samples = len(data)

    errors = []

    # check sample count in data
    if n_samples != samples_per_file:
        errors.append(f"incorrect sample number {n_samples}")

    # file not aligned with logical file boundaries
    remainder = (idx_startfile - idx_origin) % samples_per_file
    if remainder != 0.0:
        errors.append(f"misalignment with index space {remainder}")

    # check if filename is consistent with read index
    # filename corresponds to end_index (first index of next file)
    offset = get_filedate_from_sample_file_end(idx_startfile + samples_per_file, sample_rate)
    _offset = int(filepath.name.split(".")[0])

    if offset != _offset:
        errors.append(f"filename inconsistency {offset} {_offset}")

    meta: dict[str, dict[str, int | float | str | None]] = {
        "exp": {
            "name": expname,
            "version": expvers,
            "chnl": chnl,
            "sample_rate": sample_rate,
            "samples_per_file": n_samples,
            "file_secs": file_secs,
            "radar_frequency": float(mat["d_parbl"][0][PARBL_RADAR_FREQUENCY]),
            "ipp": cfv.get("ipp"),
            "tx_pulse_length": cfv.get("tx_pulse_length"),
            "rx_start": cfv.get("rx_start"),
            "rx_end": cfv.get("rx_end"),
            "tx_start": cfv.get("tx_start"),
            "tx_end": cfv.get("tx_end"),
            "cal_on": cfv.get("cal_on"),
            "cal_off": cfv.get("cal_off"),
        },
        "ts": {
            "origin": ts_origin_sec,
            "file_start": ts_endfile_sec - file_secs,
            "file_end": ts_endfile_sec,
        },
        "date": {
            "origin": str_from_ts(ts_origin_sec),
            "file_start": str_from_ts(ts_endfile_sec - file_secs),
            "file_end": str_from_ts(ts_endfile_sec),
        },
        "sample": {
            "origin": idx_origin,
            "file_start": idx_startfile,
            "file_end": (idx_startfile + samples_per_file),
        },
        "file": {"name": filepath.name, "idx": file_idx},
    }

    pointing_data = {
        "azimuth": float(mat["d_parbl"][0][PARBL_AZIMUTH]) % 360,
        "elevation": float(mat["d_parbl"][0][PARBL_ELEVATION]),
    }

    return meta, data, pointing_data, errors


####################################################################
# EISCAT PROCESS
####################################################################


def default_pad(n_pad: int, file: Path) -> None:
    """Default pad function, print inputs"""
    print(f"pad data {n_pad} samples for {file.name}")


def default_write(data: NDArray[np.int16], file: Path) -> None:
    """Default write function, print inputs"""
    print(f"write data {len(data)} samples for {file.name}")


def default_drop(data: NDArray[np.int16], file: Path) -> None:
    """Default drop function, print inputs"""
    print(f"drop data {len(data)} samples for {file.name}")


def default_pointing(sample: int, pointing_data: dict[str, float]) -> None:
    """Default pointing function, print azimuth and elevation"""
    azimuth = pointing_data["azimuth"]
    elevation = pointing_data["elevation"]
    print(f"write pointing {azimuth} {elevation}")


def default_progress(idx: int, n_files: int, period: int = 10) -> None:
    """Default progress function, print write progress"""
    if idx + 1 == n_files or idx % period == 0:
        print(f"write progress {idx+1}/{n_files}")


def eiscat_process(
    files: list[Path],
    write_data: Optional[Callable[[NDArray[np.int16], Path], None]] = None,
    pad_data: Optional[Callable[[int, Path], None]] = None,
    drop_data: Optional[Callable[[NDArray[np.int16], Path], None]] = None,
    write_pointing: Optional[Callable[[int, dict[str, float]], None]] = None,
    log_progress: Optional[Callable[[int, int, int], None]] = None,
) -> Generator[list[str], None, None]:
    """
    Process sequence of eiscat files
    """
    if write_data is None:
        write_data = default_write
    if pad_data is None:
        pad_data = default_pad
    if drop_data is None:
        drop_data = default_drop

    # sample idx of next write
    sample_write = None

    for idx, file in enumerate(files):

        if log_progress:
            log_progress(idx, len(files), 10)

        meta, data, pointing_data, errors = eiscat_load_file(file)

        sample_file_start = meta["sample"]["file_start"]

        if sample_write is None:
            sample_write = sample_file_start

        if not errors:

            # check that we are not writing old data
            if sample_file_start < sample_write:
                errors.append("attempt to overwrite data")
            # check if zero padding is needed
            n_pad = sample_file_start - sample_write
            if n_pad > 0:
                pad_data(n_pad, file)

            # write chunk to stream
            write_data(data, file)
            # write pointint data
            if write_pointing:
                write_pointing(sample_write, pointing_data)

        else:
            drop_data(data, file)
            yield errors
            # do not increment sample_write
            continue

        # increment sample_write
        sample_write = sample_file_start + meta["exp"]["samples_per_file"]

        yield errors



"""Template of a Radar station"""

import copy
import logging
from pathlib import Path
from typing import Optional

from numpy.typing import NDArray

from pyant import Beam
from pyant.types import Parameters
from radardef.components.converter_template import Converter
from radardef.components.data_loader_template import DataLoader
from radardef.components.validator_template import Validator
from radardef.types import TargetFormat


class RadarStation:
    """
    Definition of a radar station.

    Args:
        station_id: A unique string that identifies a radar station.
        transmitter: If the radar station has a transmitter.
        receiver: If the radar station has a receiver.
        lat: Geographical latitude of radar station in decimal degrees  (North+).
        lon: Geographical longitude of radar station in decimal degrees (East+).
        alt: Geographical altitude above geoid surface of radar station in meter.
        beam: The radar beam model for this station.
        beam_parameters: The parameters for the radar beam model for this station.
        ecef (optional): The ITRS coordinates of the radar station.
        ecef_lat (optional): The latitude of the ITRS coordinates of the radar station.
        ecef_lon (optional): The longitude of the ITRS coordinates of the radar station.
        ecef_alt (optional): The altitude of the ITRS coordinates of the radar station.
        min_elevation (optional): The minimum elevation the radar can measure at.
        noise_temperature (optional): The noise temperature in Kelvins intrinsic to the radar receiver.
        power (optional): The maximum power in Watts the radar transmitter can deliver.
        power_per_element (optional): power_per_element
        frequency (optional): The frequency the radar operates at.
        validator (optional): Validator to validate the correct data format from the station
        converters (optional):  List of converters that can converte from the source format to other formats
        data_loaders (optional):  Data loader object compatible with the converted files

    """

    __logger = logging.getLogger(__name__)

    @property
    def station_id(self) -> str:
        """A unique string that identifies a radar station."""
        return self.__station_id

    @property
    def transmitter(self) -> bool:
        """If the radar station has a transmitter."""
        return self.__transmitter

    @property
    def receiver(self) -> bool:
        """If the radar station has a receiver."""
        return self.__receiver

    @property
    def lat(self) -> float:
        """Geographical latitude of radar station in decimal degrees  (North+)."""
        return self.__lat

    @property
    def lon(self) -> float:
        """Geographical longitude of radar station in decimal degrees (East+)."""
        return self.__lon

    @property
    def alt(self) -> float:
        """Geographical altitude above geoid surface of radar station in meter."""
        return self.__alt

    @property
    def beam(self) -> Beam:
        """The radar beam for this station."""
        return self.__beam

    @property
    def beam_parameters(self) -> Parameters:
        """The parameters for the radar beam model for this station."""
        return self.__beam_parameters

    @property
    def ecef(self) -> NDArray | None:
        """The ITRS coordinates of the radar station."""
        return self.__ecef

    @property
    def ecef_lat(self) -> float | None:
        """The latitude of the ITRS coordinates of the radar station."""
        return self.__ecef_lat

    @property
    def ecef_lon(self) -> float | None:
        """The longitude of the ITRS coordinates of the radar station."""
        return self.__ecef_lon

    @property
    def ecef_alt(self) -> float | None:
        """The altitude of the ITRS coordinates of the radar station."""
        return self.__ecef_alt

    @property
    def min_elevation(self) -> float | None:
        """The minimum elevation the radar can measure at."""
        return self.__min_elevation

    @property
    def noise_temperature(self) -> float | None:
        """The noise temperature in Kelvins intrinsic to the radar receiver."""
        return self.__noise_temperature

    @property
    def power(self) -> float | None:
        """The maximum power in Watts the radar transmitter can deliver."""
        return self.__power

    @property
    def power_per_element(self) -> float | None:
        """Power per element."""
        return self.__power_per_element

    @property
    def frequency(self) -> float | None:
        """The frequency the radar operates at."""
        return self.__frequency

    @property
    def validator(self) -> Validator | None:
        """Validator to confirm that the"""
        return self.__validator

    def __init__(
        self,
        station_id: str,
        transmitter: bool,
        receiver: bool,
        lat: float,
        lon: float,
        alt: float,
        beam: Beam,
        beam_parameters: Parameters,
        ecef: Optional[NDArray] = None,
        ecef_lat: Optional[float] = None,
        ecef_lon: Optional[float] = None,
        ecef_alt: Optional[float] = None,
        min_elevation: Optional[float] = None,
        noise_temperature: Optional[float] = None,
        power: Optional[float] = None,
        power_per_element: Optional[float] = None,
        frequency: Optional[float] = None,
        converters: Optional[list[Converter]] = None,
        validator: Optional[Validator] = None,
        data_loaders: Optional[list[DataLoader]] = None,
    ) -> None:

        self.__station_id = station_id
        self.__transmitter = transmitter
        self.__receiver = receiver
        self.__lat = lat
        self.__lon = lon
        self.__alt = alt
        self.__beam = beam
        self.__beam_parameters = beam_parameters
        self.__ecef = ecef
        self.__ecef_lat = ecef_lat
        self.__ecef_lon = ecef_lon
        self.__ecef_alt = ecef_alt
        self.__min_elevation = min_elevation
        self.__noise_temperature = noise_temperature
        self.__power = power
        self.__power_per_element = power_per_element
        self.__frequency = frequency
        self.__validator = validator
        self.__converters: dict[TargetFormat, Converter] = dict()
        self.__data_loaders: dict[TargetFormat, DataLoader] = dict()

        if converters is not None:
            for converter in converters:
                self.add_converter(converter)

        if data_loaders is not None:
            for data_loader in data_loaders:
                self.add_data_loader(data_loader)

    def get_converters(self) -> list[Converter]:
        """Get all converters connected to this radar"""
        return list(self.__converters.values())

    def get_data_loaders(self) -> list[DataLoader]:
        """Get all data loaders connected to this radar"""
        return list(self.__data_loaders.values())

    def add_converter(self, converter: Converter) -> None:
        """Add a converter to this radar"""
        self.__converters[converter.target_format] = converter

    def add_data_loader(self, data_loader: DataLoader) -> None:
        """Add a data loader to this radar"""
        self.__data_loaders[data_loader.converted_format] = data_loader

    def convert(self, path: Path, target_format: TargetFormat, dst: Path) -> list[Path] | None:
        """Convert data from this radar to a specific format"""

        if self.__validator is not None:
            if not self.__validator.validate(path):
                self.__logger.error("Source format not supported")
                return None
        else:
            self.__logger.error("No source validator available, will try to convert anyways")

        try:
            return self.__converters[target_format].convert(path, dst)
        except KeyError:
            self.__logger.error("Not possible to convert to the wanted format, no converter available")

        return None

    def load_data(self, path: Path, converted_format: Optional[TargetFormat] = None) -> DataLoader | None:
        """Load converted data from this radar"""
        if converted_format is None:
            for data_loader in self.__data_loaders.values():
                if data_loader.validate(path):
                    loader = copy.deepcopy(data_loader)
                    loader.load(path)
                    return loader
            self.__logger.error("Source format not supported")
        else:
            try:
                if self.__data_loaders[converted_format].validate(path):
                    loader = copy.deepcopy(self.__data_loaders[converted_format])
                    loader.load(path)
                    return loader
                else:
                    self.__logger.error("Source format not supported")
            except KeyError:
                self.__logger.error("Not possible to load the wanted format, no loader available")

        return None



"""
Class and freestanding functions to convert Eiscat mat.bz2 data to DRF format, the converter is based on
the Converter template
"""

import configparser
import logging
import os
from pathlib import Path
from typing import Any, Optional

import numpy as np
import scipy.constants
from tqdm import tqdm

import radardef.radar_stations.eiscat.utils.digitalrf_wrapper as drf_wrapper
from radardef.components.converter_template import Converter
from radardef.radar_stations.eiscat.utils.eiscat_utils import (
    eiscat_files,
    eiscat_load_file,
    eiscat_process,
)
from radardef.types import Boundparam, Expparam, Metaparam, SourceFormat, TargetFormat


class EiscatMatbzToDrf(Converter):
    """Converts from Eiscat mat.bz2 format to DRF format"""

    __logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        self.__compression = 0
        super().__init__(SourceFormat.EISCAT_MATBZ, TargetFormat.DRF)

    def set_compression(self, level: int) -> None:
        """Set compression of the DRF conversion"""
        self.__compression = level

    def convert(self, src: Path, dst: Path) -> list[Path]:
        """Convert file from mat.bz2 to DRF"""
        return [
            convert_eiscat_to_drf(
                src, dst, compression=self.__compression, progress=False, logger=self.__logger
            )
        ]


def convert_eiscat_to_drf(
    src: Path,
    dst: Path,
    name: Optional[str] = None,
    compression: int = 0,
    progress: bool = False,
    logger: Optional[logging.Logger] = None,
) -> Path:
    """
    Converts Eiscat raw data to Hardtarget DRF.

    The output (Hardtarget DRF) folder will be placed within the 'dst'
    directory. By default, the name of the Hardtarget DRF folder is constructed
    from the name of the 'src' directory. The name of the Hardtarget DRF folder
    may be specified using the 'name' option. If 'name' is None, name is derived
    from 'src' (leo_bpark_2.1u_NO@uhf -> leo_bpark_2.1u_NO@uhf_drf).

    Args:

        src : Path to source directory (Eiscat raw data)
        dst : Path to destination directory
        name (optional): Name of output directory
        compression (optional): Compression level for h5 files in output, 0-9.
        progress (optional): Print download progress bar to stdout, default False
        logger (optional): logger object

    Returns:
        The absolute path to the DRF folder or None.

    Raises:
        FileNotFoundError: 'src' does not exist 'dst' does not exist
        FileExistsError: 'dst/name' already exists

    """

    src = Path(src)
    if not (src.is_dir() or src.is_file()):
        raise FileNotFoundError(str(src))

    dst = Path(dst)
    if not dst.is_dir():
        os.makedirs(dst)
        # raise FileNotFoundError(str(dst))

    if name is None:
        name = f"{src.name}_drf"
    hdrf = dst / name
    if hdrf.exists():
        raise FileExistsError(str(hdrf))
    hdrf.mkdir(parents=True, exist_ok=True)

    files = eiscat_files(src)
    n_files = len(files)

    # load experiment info from first matlab file
    meta_first = eiscat_load_file(files[0])[0]

    # create sample writer
    sample_writer = drf_wrapper.DigitalRFWriter(
        hdrf,
        meta_first["exp"]["chnl"],
        meta_first["exp"]["sample_rate"],  # sample rate numerator
        1,  # samplerate denominator
        np.int16,
        meta_first["sample"]["file_start"],
        subdir_cadence_secs=3600,  # one dir per hour
        file_cadence_secs=1,  # one file per second
        is_complex=True,
        compression_level=compression,
        uuid_str=meta_first["exp"]["chnl"],
        ts_align_sec=meta_first["ts"]["file_start"],
    )

    # create pointing writer
    pointing_writer = drf_wrapper.DigitalMetadataWriter(
        hdrf,
        "pointing",
        meta_first["exp"]["sample_rate"],  # sample rate - numerator (int)
        # sample rate - denominator (int)
        meta_first["exp"]["samples_per_file"],
    )

    def pad_data(n_pad: int, file: Path) -> None:
        """Zero pad data with n zeros."""
        try:
            sample_writer.write(np.zeros(n_pad * 2, dtype=np.int16))
        except Exception as e:
            err = f"unable to zero pad samples for {file}"
            if logger:
                logger.error(err)
            raise e
        if logger:
            logger.debug(f"zero padding {n_pad} samples for {file}")

    def write_data(data: Any, file: Path) -> None:
        """Write data to sample writer."""
        try:
            sample_writer.write(data)
        except Exception as e:
            err = f"unable to write samples for {file}"
            if logger:
                logger.error(err)
            raise e

    def drop_data(errors: Any, file: Path) -> None:
        """On data drop log error and dropped file"""
        if logger:
            logger.info(f"dropping : {errors} : {file}")

    def log_progress(idx: int, n_files: int, period: int = 10) -> None:
        """Log the convert process"""
        if logger:
            if idx + 1 == n_files or idx % period == 0:
                logger.debug(f"write progress {idx+1}/{n_files}")

    def write_pointing(sample: int, pointing_data: dict[str, float]) -> None:
        """write pointing data"""
        ts = sample_writer.ts_from_index(sample)
        pointing_idx = int(pointing_writer.index_from_ts(ts))
        pointing_writer.write(pointing_idx, pointing_data)

    if logger:
        logger.info(f"writing DRF from {n_files} input files")

    if progress:
        pbar = tqdm(desc=f"Converting files to digital_rf: {src}", total=n_files)

    # processing loop
    for error in eiscat_process(
        files,
        write_data=write_data,
        pad_data=pad_data,
        drop_data=drop_data,
        write_pointing=write_pointing,
        log_progress=log_progress,
    ):
        if error:
            if logger:
                logger.debug(error)

        if progress:
            pbar.update(1)

    if progress:
        pbar.close()

    if logger:
        logger.info("Done writing DRF files")

    meta = configparser.ConfigParser()

    # Experiment
    meta.add_section(Metaparam.EXPERIMENT)
    exp = meta[Metaparam.EXPERIMENT]

    exp[Expparam.NAME] = str(meta_first["exp"]["name"])
    exp[Expparam.TX_PULSE_LENGTH] = str(meta_first["exp"]["tx_pulse_length"])
    exp[Expparam.T_RX_START_USEC] = str(meta_first["exp"]["rx_start"])
    exp[Expparam.T_RX_END_USEC] = str(meta_first["exp"]["rx_end"])
    exp[Expparam.T_TX_START_USEC] = str(meta_first["exp"]["tx_start"])
    exp[Expparam.T_TX_END_USEC] = str(meta_first["exp"]["tx_end"])
    exp[Expparam.T_CAL_ON_USEC] = str(meta_first["exp"]["cal_on"])
    exp[Expparam.T_CAL_OFF_USEC] = str(meta_first["exp"]["cal_off"])
    exp[Expparam.RADAR_FREQUENCY] = str(meta_first["exp"]["radar_frequency"])
    exp[Expparam.RX_CHANNELS] = str([meta_first["exp"]["chnl"]])
    exp[Expparam.TX_CHANNEL] = meta_first["exp"]["chnl"]
    exp[Expparam.SAMPLE_RATE] = str(meta_first["exp"]["sample_rate"])
    exp[Expparam.T_IPP_USEC] = str(int(meta_first["exp"]["ipp"]))

    # Bounds
    meta.add_section(Metaparam.BOUNDS)
    bounds = meta[Metaparam.BOUNDS]
    meta_last = eiscat_load_file(files[-1])[0]
    bounds[Boundparam.TS_START_USEC] = str(meta_first["ts"]["file_start"])
    bounds[Boundparam.TS_END_USEC] = str(meta_last["ts"]["file_end"])

    # write metadata file
    metafile = hdrf / "metadata.ini"
    with open(metafile, "w") as f:
        meta.write(f)

    return hdrf



"""Eiscat 3D radar object"""

from pyant.beam import Beam
from pyant.types import Parameters
from radardef.components.radar_station_template import RadarStation
from radardef.radar_stations.eiscat.beams.e3d import (
    eiscat_3d_single_subarray_beam,
    eiscat_3d_stage1_beam,
    eiscat_3d_stage1_interp_beam,
    eiscat_3d_stage2_beam,
    eiscat_3d_stage2_interp_beam,
)
from radardef.types import Eiscat3DLocation, Stage, StationID


class Eiscat3D(RadarStation):
    """
    Eiscat 3D radar definition

    Args:
        location: Eiscat 3d radar location (skibotn, karesuvanto, kaiseniemi)
        stage (optional): Stage of the radar (single, stage_1, stage_2), default single
        interpolation (optional): Should interpolation be used, default False

    """

    def __init__(
        self,
        location: Eiscat3DLocation = Eiscat3DLocation.KAISENIEMI,
        stage: Stage = Stage.SINGLE,
        interpolation: bool = False,
    ):

        beam: Beam
        params: Parameters
        if interpolation:
            if stage is Stage.SINGLE:
                self.__logger.warning("There is no interpolation beam available for a single subarray")
            elif stage is Stage.STAGE_1:
                beam, params = eiscat_3d_stage1_interp_beam()
            elif stage is Stage.STAGE_2:
                beam, params = eiscat_3d_stage2_interp_beam()
            else:
                self.__logger.warning("No available beam for the given stage")
        else:
            if stage is Stage.SINGLE:
                beam, params = eiscat_3d_single_subarray_beam()
            elif stage is Stage.STAGE_1:
                beam, params = eiscat_3d_stage1_beam()
            elif stage is Stage.STAGE_2:
                beam, params = eiscat_3d_stage2_beam()
            else:
                self.__logger.warning("No available beam for the given stage")

        match location:
            case Eiscat3DLocation.SKIBOTN:
                super().__init__(
                    station_id=StationID.EISCAT_3D_SKIBOTN,
                    transmitter=True,
                    receiver=True,
                    lat=69.34023844,
                    lon=20.313166,
                    alt=0.0,
                    beam=beam,
                    beam_parameters=params,
                    min_elevation=30,
                    noise_temperature=150,
                    power=3.276e6,
                    power_per_element=1e3,
                    frequency=233e6,
                )
            case Eiscat3DLocation.KARESUVANTO:
                super().__init__(
                    station_id=StationID.EISCAT_3D_KARESUVANTO,
                    transmitter=False,
                    receiver=True,
                    lat=68.463862,
                    lon=22.458859,
                    alt=0.0,
                    beam=beam,
                    beam_parameters=params,
                    min_elevation=30,
                    noise_temperature=150,
                    frequency=233e6,
                )
            case Eiscat3DLocation.KAISENIEMI:
                super().__init__(
                    station_id=StationID.EISCAT_3D_KAISENIEMI,
                    transmitter=False,
                    receiver=True,
                    lat=68.148205,
                    lon=19.769894,
                    alt=0.0,
                    beam=beam,
                    beam_parameters=params,
                    min_elevation=30,
                    noise_temperature=150,
                    frequency=233e6,
                )
            case _:
                super().__init__(
                    station_id=StationID.UNKNOWN,
                    transmitter=False,
                    receiver=False,
                    lat=0.0,
                    lon=0.0,
                    alt=0.0,
                    beam=None,
                    beam_parameters=None,
                    min_elevation=0,
                    noise_temperature=0,
                    frequency=0,
                )



"""
This module provides a wrapper around reader and writer objects
from 'digital_rf' - to provide a uniform API and correct some design issues.

NOTE - not sure if digital_rf is thread-safe with respect to concurrent
writes to same file (e.g. adjacent blocks).
Also, it appears that digital_rf is not random access with respect to
writing (even in non-continuous mode) as it maintains and internal index
for next available sample.
"""

from pathlib import Path
from typing import Any, Iterable, Optional

import digital_rf
import numpy as np

from radardef.radar_stations.eiscat.utils.drf_utils import index_from_ts, ts_from_index


class BaseIndexedTimeSequence:
    """

    Time sequence to index and index to time sequence converter

    Args:
        sample_rate: Sample rate for the time sequence
        ts_align_sec (optional): Timestamp associated with a specific index (timestamp in seconds since epoch),
                typically it could be the timestamp of the first sample, default is 0.
                ts_align_sec defines the precise alignment between timestamps and index-domain.
        ts_offset_sec (optional): Timestamp in seconds since epoch,
                                  defines a time offset for the index space, default is 0.
    """

    def __init__(self, sample_rate: float, ts_align_sec: int = 0, ts_offset_sec: int = 0) -> None:

        self.sample_rate = sample_rate
        self.__ts_offset_sec = ts_offset_sec

        self.__ts_align_sec = ts_align_sec
        self.__ts_skew = 0
        __idx_align = index_from_ts(ts_align_sec, self.sample_rate)
        self.__ts_skew = __idx_align - np.floor(__idx_align)

    def ts_from_index(self, idx: int) -> float:
        """convert index to timestamp (seconds since epoch)"""
        ts = ts_from_index(idx, self.sample_rate, ts_offset_sec=self.__ts_offset_sec)
        return ts + self.__ts_skew

    def index_from_ts(self, ts: float) -> int:
        """Convert timestamp(seconds since epoch) to index"""
        ts = ts - self.__ts_skew
        return index_from_ts(ts, self.sample_rate, ts_offset_sec=self.__ts_offset_sec)


class DigitalRFWriter(BaseIndexedTimeSequence):
    """
    Convenience wrapper around digital_rf.DigitalRFWriter

    Args:
        dst: Destination directory.
        chnl: Radar channel.
        sample_rate_numerator: Numerator of sample rate in Hz.
        sample_rate_denominator: Denominator of sample rate in Hz.
        dtype_str: dtype of the data.
        start_global_index: The index of the first sample given in number of samples since the
            epoch.
        subdir_cadence_secs (optional): Number of seconds of data to store in one subdirectory,
            default is 3600 seconds.
        file_cadence_secs (optional): Number of seconds of data to store in each file,
            default is 3600 seconds.
        compression_level (optional): 0 for no compression (default), 1-9 for varying levels of gzip
            compression (1 == least compression, least CPU; 9 == most
            compression, most CPU).
        is_complex (optional): Used when dtype is not complex, if true the data is interpreted as complex.
        checksum (optional): If HDF5 cheksum capability should be used.
        num_subchannels (optional): Number of subchannels to write to simultaniously, default is 1.
        is_continuous (optional): If true data will be written in continuous blocks.
        marching_periods (optional): Write a period to stdout for every file when writing.
        uuid_str (optional): UUID string that will act as a unique identifier for the data. If None, a random
            UUID will be generated.
        ts_align_sec (optional): align sec

    """

    def __init__(
        self,
        dst: str | Path,
        chnl: str,
        sample_rate_numerator: float,
        sample_rate_denominator: float,
        dtype_str: Any,
        start_global_index: int,
        subdir_cadence_secs: int = 3600,  # 1 dir per hour
        file_cadence_secs: int = 3600,  # 1 hour per file
        compression_level: int = 0,
        is_complex: bool = False,
        checksum: bool = False,
        num_subchannels: int = 1,
        is_continuous: bool = True,
        marching_periods: bool = False,
        uuid_str: Optional[str] = None,
        ts_align_sec: int = 0,
    ) -> None:

        # check dst
        dst = Path(dst)
        if not dst.is_dir():
            raise Exception(f"<dst> must be directory path, {dst}")

        # channel directory
        chnldir = dst / chnl
        chnldir.mkdir(parents=True, exist_ok=True)

        # sample rate
        sample_rate = sample_rate_numerator / float(sample_rate_denominator)

        super().__init__(sample_rate, ts_align_sec=ts_align_sec)

        # meta data writer
        self._writer = digital_rf.DigitalRFWriter(
            str(chnldir),
            dtype_str,
            subdir_cadence_secs,
            file_cadence_secs * 1000,  # file_cadence_milliseconds
            start_global_index,  # start global index
            sample_rate_numerator,
            sample_rate_denominator,
            uuid_str=uuid_str,
            compression_level=compression_level,
            checksum=checksum,
            is_complex=is_complex,
            num_subchannels=num_subchannels,
            is_continuous=is_continuous,
            marching_periods=marching_periods,
        )

    def close(self) -> None:
        """Close writer"""
        self._writer.close()

    def write(self, batch: Any) -> None:
        """Write batch"""
        self._writer.rf_write(batch)


class DigitalMetadataWriter(BaseIndexedTimeSequence):
    """
    Convenience wrapper around digital_rf.DigitalMetadataWriter

    Args:
        dst: Destination directory.
        chnl: Radar channel.
        sample_rate_numerator: Numerator of sample rate in Hz.
        sample_rate_denominator: Denominator of sample rate in Hz.
        subdir_cadence_secs (optional): Number of seconds of data to store in one subdirectory,
            default is 3600 seconds.
        file_candence_secs (optional): Number of seconds of data to store in each file,
            default is 3600 seconds.
        prefix: prefix of folder
        ts_align_sec: align sec
    """

    def __init__(
        self,
        dst: str | Path,
        chnl: str,
        sample_rate_numerator: float,
        sample_rate_denominator: float,
        subdir_cadence_secs: int = 3600,  # 1 dir per hour
        file_candence_secs: int = 3600,  # 1 hour per file
        prefix: str = "meta",
        ts_align_sec: int = 0,
    ) -> None:

        # check dst
        path = Path(dst)
        if not path.is_dir():
            raise Exception(f"<dst> must be directory path, {path}")

        # metadir
        metadir = path / chnl
        metadir.mkdir(parents=True, exist_ok=True)

        # sample rate
        sample_rate = sample_rate_numerator / float(sample_rate_denominator)

        super().__init__(sample_rate, ts_align_sec=ts_align_sec)

        # meta data writer
        self._writer = digital_rf.DigitalMetadataWriter(
            str(metadir),
            subdir_cadence_secs,
            file_candence_secs,
            sample_rate_numerator,
            sample_rate_denominator,
            prefix,
        )

    def write(self, idx: int, data: dict[str, Any]) -> None:
        """Write data at sample"""
        self._writer.write(idx, data)

    def close(self) -> None:
        """Close writer"""
        pass



"""
Format validator gathers all given radars and extracts their source validators, the collection can then solve
what format any given file is, given that the format is the source format of atleast one radar.
"""

import logging
from collections.abc import Callable
from pathlib import Path

from radardef.components.radar_station_template import RadarStation
from radardef.types import SourceFormat


class FormatCollection:
    """
    Collection of all available validators from the available radar objects,
    provides a way to gather and provide source validators for many different formats in one place and make
    it possible to solve the source format of a given file.

    Args:
        radars: List of radars that the collections validators will be extracted from.

    """

    __logger = logging.getLogger(__name__)

    def __init__(self, radars: list[RadarStation]) -> None:

        self.__formats: dict[SourceFormat, Callable[[Path], bool]] = dict()

        for radar in radars:
            if radar.validator is not None:
                self.__logger.info(f"registering  {radar.validator}")
                self._register_validator(radar.validator.format, radar.validator.validate)

    def _register_validator(self, source_format: SourceFormat, validator: Callable) -> None:
        """Register a validator related to a source format to collection."""
        self.__formats[source_format] = validator

    def get_format(self, path: Path) -> SourceFormat:
        """Returns the source format of the file."""

        for source_format, validator in self.__formats.items():
            if validator(path):
                return source_format
        return SourceFormat.UNKNOWN

    def is_format(self, path: Path, source_format: SourceFormat) -> bool:
        """Validates that the file is of the expected format"""
        try:
            return self.__formats[source_format](path)
        except KeyError:
            self.__logger.debug(f"No validator available for {source_format} format")
            return False

    def list_formats(self) -> str:
        """Lists formats with a validator as a string"""

        st = ""
        if self.__formats is not None:
            for source_format, _ in self.__formats.items():
                st += f">{source_format}\n"
        return st



__all__ = [
    "Mu",
    "EiscatUHF",
    "Eiscat3D",
    "EiscatVHF",
    "ESR",
    "TSDR",
    "Pansy",
]

from radardef.radar_stations.eiscat import ESR, TSDR, Eiscat3D, EiscatUHF, EiscatVHF
from radardef.radar_stations.mu import Mu
from radardef.radar_stations.pansy import Pansy



"""Enums and typed dicts to describe metadata, a standardized dict that each data loader should return"""

from typing import NamedTuple, Optional

import numpy as np
import numpy.typing as npt


# TODO: SI units?
class ExpParams(NamedTuple):
    """
    Experiment parameters
    """

    name: str = "Default"
    radar_frequency: float = 0.0
    t_ipp_usec: float = 0
    ipp_samps: int = 0
    sample_rate: float = 0.0
    t_samp_usec: float = 0
    rx_channels: list[str] | list[int] = []
    t_rx_start_usec: float = 0.0
    t_rx_end_usec: float = 0.0
    t_tx_start_usec: float = 0.0
    t_tx_end_usec: float = 0.0
    wavelength: float = 0.0
    tx_channel: Optional[str | int] = None
    tx_pulse_length: Optional[int] = None
    t_cal_on_usec: Optional[float] = None
    t_cal_off_usec: Optional[float] = None
    data: Optional[np.datetime64] = None
    code: npt.NDArray[np.float64] = np.empty((10,), dtype=np.float64)
    pulse: Optional[int] = None


class BoundParams(NamedTuple):
    """
    Time since epoch (standard unix) bounds
    """

    ts_start_usec: float | int = 0
    ts_end_usec: float | int = 0


class Metadata(NamedTuple):
    """
    Metadata
    """

    experiment: ExpParams
    bounds: BoundParams


class Pointing(NamedTuple):
    """
    Pointing
    """

    data: list[tuple[int, dict[str, float]]]
    sample_rate: float



"""
Class and freestanding functions to load metadata and data from a file converted from Eiscat mat.bz2
to DRF format. The data loader is based on the DataLoader template.
"""

import configparser
from pathlib import Path
from typing import Optional

import digital_rf
import numpy as np
import numpy.typing as npt
import scipy.constants

from radardef.components.data_loader_template import DataLoader
from radardef.radar_stations.eiscat.utils import load_radar_code
from radardef.radar_stations.eiscat.validators import DRF
from radardef.types import (
    Boundparam,
    BoundParams,
    Expparam,
    ExpParams,
    Metadata,
    Metaparam,
    Pointing,
    TargetFormat,
)


class DrfLoader(DataLoader):
    """Simplifies the way to load DRF files converted from eiscat"""

    @property
    def meta(self) -> Metadata:
        """Metadata, containing experiment data and bounds data"""

        # metadata file
        meta_file = configparser.ConfigParser()
        meta_file.read(self.__path / "metadata.ini")
        sample_rate = meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.SAMPLE_RATE)
        t_samp_usec = int((1 / sample_rate) * 1e6)
        t_ipp_usec = meta_file.getint(Metaparam.EXPERIMENT, Expparam.T_IPP_USEC)
        ipp_samps = int(t_ipp_usec * 1e-6 * sample_rate)
        code = load_radar_code("leo_bpark")  # TODO: Base this on some experiment param in the future

        experiment = ExpParams(
            name=meta_file.get(Metaparam.EXPERIMENT, Expparam.NAME).strip("'").strip('"'),
            radar_frequency=meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.RADAR_FREQUENCY),
            t_ipp_usec=t_ipp_usec,
            sample_rate=sample_rate,
            ipp_samps=ipp_samps,
            t_samp_usec=t_samp_usec,
            rx_channels=eval(meta_file.get(Metaparam.EXPERIMENT, Expparam.RX_CHANNELS).strip("'").strip('"')),
            tx_channel=meta_file.get(Metaparam.EXPERIMENT, Expparam.TX_CHANNEL).strip("'").strip('"'),
            tx_pulse_length=int((meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.TX_PULSE_LENGTH) + 1)),
            t_rx_start_usec=meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.T_RX_START_USEC),
            t_rx_end_usec=meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.T_RX_END_USEC),
            t_tx_start_usec=meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.T_TX_START_USEC),
            t_tx_end_usec=(meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.T_TX_END_USEC) + t_samp_usec),
            t_cal_on_usec=meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.T_CAL_ON_USEC),
            t_cal_off_usec=meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.T_CAL_OFF_USEC),
            wavelength=scipy.constants.c
            / (meta_file.getfloat(Metaparam.EXPERIMENT, Expparam.RADAR_FREQUENCY) * 1e6),
            code=code,
        )

        bounds = BoundParams(
            ts_start_usec=meta_file.getfloat(Metaparam.BOUNDS, Boundparam.TS_START_USEC),
            ts_end_usec=meta_file.getfloat(Metaparam.BOUNDS, Boundparam.TS_END_USEC),
        )

        return Metadata(experiment, bounds)

    @property
    def pointing(self) -> Pointing:
        """Pointing data, data describing the radar pointing direction in spherical coordinates"""
        if self.__meta_reader is None:
            return Pointing(data=[(0, {"azimuth": 0, "elevation": 0})], sample_rate=0)
        else:
            idx_start, idx_end = self.__meta_reader.get_bounds()
            # load pointing data as vector
            return Pointing(
                data=list(self.__meta_reader.read(idx_start, idx_end).items()),
                sample_rate=float(self.__meta_reader.get_samples_per_second()),
            )

    @property
    def channels(self) -> list[int] | list[str]:
        """All available channels"""
        return self.__channel_reader.get_channels()

    def __init__(self) -> None:
        super().__init__(TargetFormat.DRF, DRF())

    def load(self, path: Path | str) -> None:
        """Loads a path to the dataloader, extracting metadata and other important specifications

        Args:
            path: path to data file

        Raises:
            Exception: path is not a directory
            Exception: pointing dir must be a directory
        """

        self.__path = Path(path).resolve()

        if not self.__path.is_dir():
            raise Exception(f"<dir> must be directory path, {self.__path}")

        self.__channel_reader = digital_rf.DigitalRFReader(str(self.__path))

        pointing_dir = self.__path / "pointing"
        if not pointing_dir.is_dir():
            raise Exception(f"<dir/pointing> must be directory path, {self.__path}")

        self.__meta_reader = digital_rf.DigitalMetadataReader(str(pointing_dir))

    def bounds(self, channel: str | int) -> tuple[int, int]:
        """Sample bounds of the specific channel

        Args:
            channel: channel
        Returns:
            tuple of start and end sample
        Raises:
            Exception: channel is missing
        """

        if not isinstance(channel, str):
            channel = str(channel)

        if channel not in self.__channel_reader.get_channels():
            raise Exception(f"channel {channel} missing in {dir}")

        idx_first, idx_last = self.__channel_reader.get_bounds(channel)
        return idx_first, idx_last + 1

    def read(
        self,
        channel: str | int,
        start_sample: Optional[int] = None,
        vector_length: Optional[int] = None,
    ) -> npt.NDArray[np.complex128]:
        """
        Read data from loaded file

        Args:
            channel (optional): Channel to read data from
            start_sample (optional): Start of range to read, if empty all data will be read
            vector_length (optional): Number of samples (counting from start sample) to read,
                                    if empty all data will be read

        Returns:
            Complex data of length vector_length from give channel

        Raises:
            Exception: channel is missing
        """

        if not isinstance(channel, str):
            channel = str(channel)

        if channel not in self.__channel_reader.get_channels():
            raise Exception(f"channel {channel} missing in {dir}")

        if start_sample is None or vector_length is None:
            bound_start, bound_end = self.bounds(channel)
            return self.__channel_reader.read_vector_1d(bound_start, bound_end - 1, channel)
        else:
            return self.__channel_reader.read_vector_1d(start_sample, vector_length, channel)



from .formats import (
    BeamType,
    Boundparam,
    DishDiameter,
    Eiscat3DLocation,
    EiscatUHFLocation,
    Expparam,
    Metaparam,
    SourceFormat,
    Stage,
    StationID,
    TargetFormat,
)
from .types import BoundParams, ExpParams, Metadata, Pointing



"""
This module contains functionality to convert MUI files to h5 for swifter loading and better data overview.
The converter is based on the Converter template

Conventions of the h5-format
-----------------------------

* dates and time are saved in numpy.datetime64[ns]
* the voltage data nd-array is saved in a dataset called "/data"

"""

import logging
import os
import warnings
from io import BufferedReader
from pathlib import Path
from typing import Any

import h5py
import numpy as np
from tqdm import tqdm

import radardef.tools as tools
from radardef.components.converter_template import Converter
from radardef.types import SourceFormat, TargetFormat

logger = logging.getLogger(__name__)


class MuiToH5(Converter):
    """MUI to H5 converter."""

    logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        super().__init__(SourceFormat.MUI, TargetFormat.H5)

    def convert(self, src: Path, dst: Path) -> list[Path]:
        """Convert MUI to h5, converted files are saved in dst"""
        output_directories: list[Path] = []
        if src.is_dir():
            for file in tqdm(
                src.iterdir(),
                total=len([f for f in src.iterdir() if f.is_file()]),
                desc=f"Converting files to h5: {src}",
            ):
                output_directories.append(convert_mui_to_h5(file, dst))
        elif src.is_file():
            output_directories.append(convert_mui_to_h5(src, dst))
        return list(set(output_directories))


def convert_mui_to_h5(
    src: Path, dst: Path, experiment_name: str = "mw26x6", skip_existing: bool = False
) -> Path:
    """
    Converts MU raw data to HDF5.

    Args:
        src: Path to source directory (Eiscat raw data)
        dst: Path to destination directory
        experiment_name (optional): Experiment name
        skip_existing (optional): Skip already existing files

    Returns:
        Path to generated HDF5 file directory

    """

    file_outputs_created = []

    try:
        file = open(src, "rb")
    except TypeError:
        logger.debug("File already open or wrong input type.")

    if hasattr(file, "mode"):
        if file.mode != "rb":
            logger.critical("File not open in binary mode.")
            return Path("")

    """
    Constants declared
    """
    SKIP_AMOUNT = 4480

    byte = file.read(1)
    file.seek(-1, 1)
    while byte != b"":
        """
        Initiating variables that will be used later.
        """
        header_data = _get_header_data(file)
        block_amount = header_data["mu_head_1_to_24"][2]
        observation_param_name = header_data["observation_param_name"]
        # I replace ':' with a ., as windows cannot save files with ':' in their name.
        start_time = np.datetime_as_string(header_data["record_start_time"]).replace(":", ".")
        dst_dated = (
            Path(str(dst)).joinpath(
                start_time[0:4],  # Year
                start_time[5:7],  # Month
                start_time[8:10],  # Day
            )
            if bool(dst)
            else Path("")
        )
        """
        Combined it will be something like "dst/20XX/YY/ZZ"
        Using str(dst) to safeguard against dst being None
        """
        output_file_name = Path(dst_dated).joinpath(start_time + ".h5")

        """
        Creates the directories if not yet created.
        """
        if not dst_dated.name == "" and not Path(dst_dated).is_dir():
            logger.debug(f"Creating file directories for location {dst_dated}")
            os.makedirs(dst_dated)

        if bool(dst):
            """
            Checks if dst has been set and if the directory already exists.
            If it does, it will check if the output file already exists. If it does and
            the "skip existing files" flag is set, it will skip those datasets.
            """
            # If the file already exists and you want to skip it
            if Path(output_file_name).is_file() and skip_existing:
                logger.debug(f'Skip existing set and file was found. Skipping file "{output_file_name}".')
                # Skip ahead to the next block
                file.seek(block_amount * SKIP_AMOUNT, 1)
                byte = file.read(1)
                file.seek(-1, 1)
                continue

        if observation_param_name.strip() != experiment_name:
            logger.critical(
                f'Experiment name "{experiment_name}" was not '
                + f'equal to observation parameter name "{observation_param_name}". Exiting.'
            )
            break

        """
        Opening the file properly so it closes if it crashes.
        """
        with h5py.File(output_file_name, "w") as h5file:
            """
            Allocating variables to hold the data.
            mu_data = [channel][ipp][sample]
            """
            ipp_samples = 85
            n_ipp = 512
            channels = 25
            channel_list = []
            mu_data = np.zeros((channels, n_ipp, ipp_samples), dtype="complex")

            for x in range(0, block_amount):
                """
                Block structure:

                    | 8 bit|  8 bit  | 16 bit |   32 bit x 512   |     32 bit x 512      |
                    | beam | channel | sample | 512 real numbers | 512 imaginary numbers |

                    Each block represents the x sample in each ipp for channel x,
                    when all samples for a channel has been read the next channel begins.
                    can be represented like:

                    block: 1    | sample 1 for ipp 1-512 in channel 1|
                    block: 2    | sample 2 for ipp 1-512 in channel 1|
                    .......
                    block: 85   | sample 85 for ipp 1-512 in channel 1|
                    block: 86   | sample 1 for ipp 1-512 in channel 2|


                NOTE: each block ends with 380 blank bytes and should be skipped
                """

                beam = np.fromfile(file, dtype=">i1", count=1)[0]
                channel = np.fromfile(file, dtype=">i1", count=1)[0]  # alt: (x // ipp_samples) % channels
                ipp_sample = np.fromfile(file, dtype=">i2", count=1)[0]  # x % ipp_samples

                real_numbers = np.fromfile(file, dtype=">f4", count=n_ipp)
                complex_numbers = np.fromfile(file, dtype=">f4", count=n_ipp)
                data = real_numbers + 1j * complex_numbers

                mu_data[channel - 1, :, ipp_sample - 1] = data

                if channel not in channel_list:
                    channel_list.append(channel)
                file.seek(380, 1)

            """
            Adding all header data as attributes to the hdf5 file.
            """
            for key, val in header_data.items():
                logger.debug(f"Setting file attribute {key} to {val}")
                if type(val) == np.datetime64:
                    h5file.attrs[key] = str(np.datetime_as_string(val))
                else:
                    h5file.attrs[key] = val
            h5file.attrs["filename"] = src.name
            h5file.attrs["date"] = str(np.datetime_as_string(header_data["record_start_time"]))
            h5file.attrs["path"] = str(src.resolve())

            logger.debug("Creating dataset data, and saving to file")
            h5file.create_dataset("data", data=mu_data)
            h5file.create_dataset("rx_channels", data=channel_list)
            file_outputs_created.append(str(output_file_name))

        byte = file.read(1)
        file.seek(-1, 1)

    logger.debug("Reached EOF, exiting loop and closing file")
    file.close()
    return dst_dated


def _get_header_data(file: BufferedReader) -> dict:
    """
    Retrieves the meta/headerdata from a MUI file for later conversion and for
    use in parsing information.
    """
    header_data: dict[str, Any] = dict()

    # Copyright Csilla Szasz, Kenneth Kullbrandt, Daniel Kastinen
    """
    %-- Row 1 of the header
    %-- Reading the header of one record
    %-- (Byte position: 1-24) The first 6 numbers are integers,
    %-- four bytes long, each byte consisting of 8 bits -> each
    %-- integer is represented by 4x8=32 bits.
    """
    header_data["mu_head_1_to_24"] = np.fromfile(file, dtype=">i4", count=6)

    """
    %-- (25-32) The data taking program name is 8 bytes
    %-- -> 8x8=64 bits and is written in ASCII.
    """
    header_data["program_name"] = _decode_utf(np.fromfile(file, dtype="S8", count=1))

    """
    %-- (33-56) Parameter file load time: 24 bytes, of characters: dd-mmm-yyyy hh:mm:ss(.ss?)
    """

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        header_data["file_load_time"] = np.datetime64(
            _convert_date(np.fromfile(file, dtype="S24", count=1)).strip() + "Z", "ns"
        )

    """
    %-- (57-60) The data taking program number, 4 bytes -> 4*8=32 bits.
    """
    header_data["program_number"] = np.fromfile(file, dtype=">i4", count=1)

    """
    %-- (61-84) Record start time: DD-MMM-YYYY hh:mm:ss.ss is 24 bytes long, each 8 bits
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        header_data["record_start_time"] = np.datetime64(
            _convert_date(np.fromfile(file, dtype="S24", count=1)).strip() + "Z", "ns"
        )

    """
    %-- (85-96) Record end time: hh:mm:ss.ss is 12 bytes long
    %-- Since "record_end_time" is only given in hh:mm:ss.ss, we want to use the 'dd-mmm-yyy ' part
    %-- of "record_start_time" to be able to convert recend into matlab serial date number.
    %-- To do this we need to convert it to a string and extract the first 10 symbsols and then
    %-- decode the rest from the file. After this we check so an edge case where the day rolled over
    %-- hasn't occured.
    """
    start_date = np.datetime_as_string(header_data["record_start_time"])[0:10]
    end_time = _decode_utf(np.fromfile(file, dtype="S12", count=1))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        temp_end_time = np.datetime64((start_date + " " + end_time).strip() + "Z", "ns")
    header_data["record_end_time"] = _fix_date_edge_case(header_data["record_start_time"], temp_end_time)

    """
    %-- (97-172) The next 19 numbers are 4 byte integers, i.e. 4*8=32 bits long
    """
    header_data["mu_head_25_to_43"] = np.fromfile(file, dtype=">i4", count=19)

    """
    %-- (173-256) The lag number of each ACF point is 21 unsigned longwords represented by 84 bytes,
    %-- i.e. each longword is 84/21=4 bytes of 8 bits each, i.e., the same thing as matlab's int32
    """
    header_data["lag_number_per_ACF"] = np.fromfile(file, dtype=">u4", count=21)

    """
    %-- (257-320) The beam direction number in first 16 beams is represented by 16 unsigned longwords
    %-- of 4 bytes and 4*8=32 bits each
    """
    header_data["beam_direction_number"] = np.fromfile(file, dtype=">u4", count=16)

    """
    %-- (321-336) The next 4 parameters are 4 byte, 4*8=32 bit, integers
    """
    header_data["mu_head_81_to_84"] = np.fromfile(file, dtype=">i4", count=4)

    """
    %-- (337-344) The HP parameter file name is 8 bytes, i.e. 8*8=64 bits
    %-- cannot add characters to the 'head' matrix
    """
    header_data["hp_file_name"] = _decode_utf(np.fromfile(file, dtype="S8", count=1))

    """
    %-- (345-352) The next two parameters are 4 byte, i.e. 4*8=32 bit, integers
    """
    header_data["mu_head_87_to_88"] = np.fromfile(file, dtype=">i4", count=2)

    """
    %-- (353-416) The number of sum (ACF method): 16 unsigned longwords -> 16 * 4 byte (4*8=32 bit) integers
    """
    header_data["number_of_sum_ACF"] = np.fromfile(file, dtype=">u4", count=16)

    """
    %-- (417-420) Number of sample poinst: 4 bytes = 32 bits
    """
    header_data["sample_points"] = np.fromfile(file, dtype=">i4", count=1)

    """
    %-- (421-600) The next 180 bytes are empty, reserved for the future, so we step
    %-- (601-616) Observation parameter name: 16 bytes
    """
    header_data["observation_param_name"] = _decode_utf(np.fromfile(file, dtype="S16", count=1, offset=180))

    """
    %-- (617-640) The next 6 parameters are 4 bytes (4*8=32bits) each
    """
    header_data["mu_head_155_to_160"] = np.fromfile(file, dtype=">i4", count=6)

    """
    %-- (641-896) Transmitted pulse pattern: 32 bits * 64
    """
    header_data["transmitted_pulse_pattern"] = np.fromfile(file, dtype=">i4", count=64)

    """
    %-- (897-904) The next 2 parameters are 4 bytes (4*8=32 bits) each
    """
    header_data["mu_head_224_to_225"] = np.fromfile(file, dtype=">i4", count=2)

    """
    %-- (905-1160) Pulse decoding pattern for all channel: 32 bits * 64
    """
    header_data["pulse_decoding_patterns"] = np.fromfile(file, dtype=">i4", count=64)

    """
    %-- (1161-1164) Beam stearing interval: 4 bytes (4*8=32 bits)
    """
    header_data["beam_stearing_interval"] = np.fromfile(file, dtype=">i4", count=1)

    """
    %-- (1165-1676) Beam direction: 2 bytes (2*8=16 bits) * 256
    %-- Cannot put this into the matrix - 256 numbers into 128 slots
    """
    header_data["beam_directions"] = np.fromfile(file, dtype=">i2", count=256)

    """
    %-- (1677-1696) The next 5 parameters are 4 bytes (2*8=32 bits) each
    """
    header_data["mu_head_420_to_424"] = np.fromfile(file, dtype=">i4", count=5)

    """
    %-- (1697-1716) TX frequency offset of 5 unsigned longwords -> 5 * 4 byte (4*8=32 bit) integers
    """
    header_data["tx_frequency"] = np.fromfile(file, dtype=">u4", count=5)

    """
    %-- (1717-1724) The next 2 parameters are 4 bytes (2*8=32 bits) each
    """
    header_data["mu_head_430_to_431"] = np.fromfile(file, dtype=">i4", count=2)

    """
    %-- (1725-1740) The RX attenuator is 4 unsigned longwords -> 4 * 4 byte (4*8=32 bit) integers
    """
    header_data["rx_attentuator"] = np.fromfile(file, dtype=">u4", count=4)

    """
    %-- (1741-1744) The state of the TX is given by a 4 byte (4*8=32 bit) integer
    """
    header_data["tx_state"] = np.fromfile(file, dtype=">i4", count=1)

    """
    %-- (1745-1760) The state of the RX is given by 4 unsigned longwords -> 4 * 4 byte (4*8=32 bit) integers
    """
    header_data["rx_state"] = np.fromfile(file, dtype=">u4", count=4)

    """
    %-- (1761:1812) The RX module selection is given by 2 bytes * (25+1)
    %-- -> cannot be saved in the head matrix (no room for it)
    """
    header_data["rx_module_selection"] = np.fromfile(file, dtype=">i2", count=26)

    """
    %-- (1813:1820) The next 2 parameters are 4 bytes (2*8=32 bits) each
    """
    header_data["mu_head_454_to_455"] = np.fromfile(file, dtype=">i4", count=2)

    """
    %-- (1821-2844) The sample start time is represented by 256
    %-- unsigned longwords (unit of sub-pulse/4) -> 256 * 4 byte (4*8=32 bit) integer
    """
    header_data["sample_start_time"] = np.fromfile(file, dtype=">u4", count=256)

    """
    %-- (2845-2848) The reception sequence: 4 byte (4*8=32 bit) integer
    """
    header_data["reception_sequence"] = np.fromfile(file, dtype=">i4", count=1)

    """
    %-- (2849-2964) The channel number in digital combine: 29 * 4 bytes (4*4=32 bits)
    """
    header_data["channel_number_digital"] = np.fromfile(file, dtype=">i4", count=29)

    """
    %-- (2965-3080) Number of coherent integrations for each combined channel:
    %-- 29 unsigned longwords -> 29 * 4 byte (4*8=32 bit) integers
    """
    header_data["coherent_integrations"] = np.fromfile(file, dtype=">u4", count=29)

    """
    %-- (3081-3196) Number of FFT points for each combined channel:
    %-- 29 unsigned longwords -> 29 * 4 byte (4*8=32 bit) integers
    """
    header_data["fft_points"] = np.fromfile(file, dtype=">u4", count=29)

    """
    %-- (3197-3312) Number of data points for each combined channel:
    %-- 29 unsigned longwords -> 29 * 4 byte (4*8=32 bit) integers
    """
    header_data["data_points"] = np.fromfile(file, dtype=">u4", count=29)

    """
    %-- (3313-3428) Lower and upper boundary of FFT number in each combined channel:
    %-- 2 * 29 * 2 bytes (2*8=16 bits)
    """
    header_data["lo_hi_bound_fft_0"] = np.fromfile(file, dtype=">i2", count=58)

    """
    %-- (3429-3544) Lower and upper boundary of FFT number in each combined channel:
    %-- 2 * 29 * 2 bytes (2*8=16 bits)
    """
    header_data["lo_hi_bound_fft_1"] = np.fromfile(file, dtype=">i2", count=58)

    """
    %-- (3545-3660) Lower and upper boundary of FFT number in each combined channel:
    %-- 2 * 29 * 2 bytes (2*8=16 bits)
    """
    header_data["lo_hi_bound_fft_2"] = np.fromfile(file, dtype=">i2", count=58)

    """
    %-- The above numbers don't fit into the head matrix (58 numbers but 29 slots)
    %-- They are put into a matrix of their own: IFFT = [IFFT1  IFFT2  IFFT3]
    %-- (3661-3776) RX frequency offset for ach combined channel:
    %-- 29 unsigned longwords -> 29 * 4 byte (4*8=32 bit) integers
    """
    header_data["rx_frequency_offset"] = np.fromfile(file, dtype=">u4", count=29)

    """
    %-- (3777-3808) FIR coefficient in RX: 16 * 2 bytes (2*8=16 bits)
    %-- Doesn't fit into the head matrix: 16 numbers into 8 slots...
    """
    header_data["fir_rx_coefficient"] = np.fromfile(file, dtype=">i2", count=16)

    """
    %-- (3809-3924) Gain adjustment of FIR filter in RX for each combined channel:
    %-- 29 * 2 bytes (2*8=16 bits each)
    %-- Cannot fit into the head matrix: 58 number into 29 slots
    """
    header_data["fir_rx_gain"] = np.fromfile(file, dtype=">i2", count=58)

    """
    %-- (3925-3940) The next four parameters are 4 bytes each, i.e. 4*8=32 bits
    """
    header_data["mu_head_982_to_985"] = np.fromfile(file, dtype=">i4", count=4)

    """
    %-- (3941-4056) Number of CIC filter in RX for each cmobined channel:
    %-- 29 unsigned longwords -> 29 * 4 byte (4*8=32 bit) integers
    """
    header_data["cic_rx_filter_amount"] = np.fromfile(file, dtype=">u4", count=29)

    """
    %-- (4057-4172) CIC cropping rate in RX for each combined channel:
    %-- 29 unsigned longwords -> 29 * 4 byte (4*8=32 bit) integers
    """
    header_data["cic_rx_cropping_rate"] = np.fromfile(file, dtype=">u4", count=29)

    """
    %-- (4173-4176) Above sea level:
    """
    header_data["above_sea_level"] = np.fromfile(file, dtype=np.float32, count=1)

    """
    %-- (4177-4180) Header flag: 4 bytes, i.e., 4*8=32 bits
    """
    header_data["header_flag"] = np.fromfile(file, dtype=">i4", count=1)

    """
    %-- (4181-4260) Comment by user: 80 bytes (80*8=640 bits)
    """
    header_data["user_comment"] = _decode_utf(np.fromfile(file, dtype="S80", count=1))

    """
    %-- (4261-4272) The next 3 parameters are 4 bytes (4*8=32 bits) each
    """
    header_data["mu_head_1066_to_1068"] = np.fromfile(file, dtype=">i4", count=3)

    """
    %-- (4273-4480) User header: 4480-4273+1 = 208 bytes
    """
    header_data["user_header"] = _decode_utf(np.fromfile(file, dtype="S208", count=1))
    return header_data


def _decode_utf(ucs_nd: np.ndarray) -> str:
    """
    Converts a UCS-4 numpy ndarray into a UTF-8 python string.
    It first views it as a chararray to access the decode method,
    then squeezes it and converts it from a ndarray into a python string.
    After this, I remove the ' at the beginning and end of the string.
    """
    return np.array2string(ucs_nd.view(np.char.chararray).decode("utf-8").squeeze())[1:-1]


def _convert_date(date_np: np.ndarray) -> str:
    """
    Since the datetimes involved are given in the form "DD-MMM-YYYY hh:mm:ss.ss"
    and numpy expects the form "YYYY-MM-DD hh:mm:ss.ss(...)", we must first convert
    the MU date to numpy standard format. This is accomplished with the help of the
    month dictionary and the _decode_utf function.
    """
    if type(date_np) == np.ndarray:
        date_str = _decode_utf(date_np)
    return date_str[7:11] + "-" + _month[date_str[3:6]] + "-" + date_str[0:2] + date_str[11:]


_month = {
    "JAN": "01",
    "FEB": "02",
    "MAR": "03",
    "APR": "04",
    "MAY": "05",
    "JUN": "06",
    "JUL": "07",
    "AUG": "08",
    "SEP": "09",
    "OCT": "10",
    "NOV": "11",
    "DEC": "12",
}


def _fix_date_edge_case(start_time: np.datetime64, end_time: np.datetime64) -> np.datetime64:
    """
    Checks so the date hasn't rolled over, and if it has it adds a day to the counter.
    """
    if end_time - start_time < 0:
        logger.info("Date rolled over for dataset")
        return end_time + np.timedelta64(1, "D")
    return end_time



__all__ = ["EiscatUHF", "Eiscat3D", "EiscatVHF", "ESR", "TSDR"]


from .eiscat_3d import Eiscat3D
from .eiscat_uhf import EiscatUHF
from .eiscat_vhf import EiscatVHF
from .esr import ESR
from .tsdr import TSDR
from .utils import load_radar_code



"""A collection of functions and information for the PANSY Radar system."""

import pathlib
from pathlib import Path
from typing import Any, Optional

import numpy as np
import numpy.typing as npt
import scipy.interpolate

from pyant.models import Array, ArrayParams, InterpolatedArray, InterpolatedArrayParams
from radardef.tools.types import CarthesianCoordinates_3xN, NDArray_2, NDArray_2xN
from spacecoords.spherical import cart_to_sph

from .data import DATA_PATHS


def load_pos_and_gain_from_data() -> tuple[npt.NDArray, npt.NDArray]:
    """Load position and gain data from the measurement files"""

    assert "antpos.csv-20160121" in DATA_PATHS, "pos file missing!"
    # SerialNo. , GrpName, ModuleID, Ready , X(m), Y(m), Z(m)
    _pansy_antennas = np.genfromtxt(
        fname=DATA_PATHS["antpos.csv-20160121"],
        skip_header=1,
        delimiter=",",
        dtype="i4, U3, U3, i4, f8, f8, f8",
    )

    assert "mu_yagi_gain.npz" in DATA_PATHS, "gain file missing!"
    _mu_yagi = np.load(DATA_PATHS["mu_yagi_gain.npz"])

    return _pansy_antennas, _mu_yagi


def pansy_array_beam() -> tuple[Array, ArrayParams]:
    """A pansy array beam"""

    _pansy_antennas, _mu_yagi = load_pos_and_gain_from_data()

    az = _mu_yagi["az_deg"].reshape(-1, 721)
    az -= 180
    el = _mu_yagi["el_deg"].reshape(-1, 721)
    gain_dB = _mu_yagi["gain_dB"].reshape(-1, 721)
    gain_dB = gain_dB - np.max(_mu_yagi["gain_dB"])

    interp = scipy.interpolate.RegularGridInterpolator(
        (az[0, :], el[:, 0]),
        gain_dB.T,
        bounds_error=False,
    )

    def yagi(cart_coord: CarthesianCoordinates_3xN, polarization: NDArray_2) -> NDArray_2xN:
        sph = cart_to_sph(cart_coord, degrees=True)
        G = 10 ** (interp(sph[:2, :].T) / 10.0)
        return np.stack([G, G], axis=0)

    groups_f1 = np.unique(_pansy_antennas["f1"])
    grp_remove = ["DIV", "Z1", "Z2"]
    groups = [grp for grp in groups_f1 if grp not in grp_remove]

    antennas = []
    for ind, grp in enumerate(groups):
        select = np.logical_and(
            _pansy_antennas["f1"] == grp,
            _pansy_antennas["f3"] == 1,
        )
        subgroup = np.empty((3, np.sum(select)), dtype=np.float64)
        subgroup[0, :] = _pansy_antennas["f4"][select]
        subgroup[1, :] = _pansy_antennas["f5"][select]
        subgroup[2, :] = _pansy_antennas["f6"][select]
        antennas.append(subgroup)

    beam = Array(
        antennas=antennas,  # type: ignore[arg-type]
        antenna_element=yagi,
    )
    params = ArrayParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=47e6,
        polarization=beam.polarization.copy(),
    )
    return beam, params


def pansy_interpolated_array_beam(
    path: Optional[Path] = None, **interpolation_kwargs: Any
) -> tuple[InterpolatedArray, InterpolatedArrayParams]:
    """
    A pansy interpolated array beam

    Beam can be loaded from a file if a path is given.
    """

    beam = InterpolatedArray()
    params = InterpolatedArrayParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
    )
    if path is None:
        array, arr_params = pansy_array_beam()
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        return beam, params

    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    if path.is_file():
        beam.load(path)
    else:
        array, arr_params = pansy_array_beam()
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        beam.save(path)
    return beam, params



"""Utility function for accessing DRF."""

import datetime as dt


def ts_from_str(datetime_str: str, as_local: bool = False) -> float:
    """
    Convert from human-readable string (ISO 8601 without a time zone) to a timestamp (seconds since epoch).

    Args:
        datetime_str: datetime as human readable string
        as_local (optional): By default, the string is interpreted as UTC time unless <as_local> is True,
            in which case the string is interpreted as local time.

    Returns:
        Timestamp
    """
    # Parse the string into a naive datetime object
    _datetime = dt.datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")

    if as_local:
        # Make it timezone-aware as local time
        _datetime = _datetime.astimezone()
    else:
        # Make it timezone-aware as UTC
        _datetime = _datetime.replace(tzinfo=dt.timezone.utc)

    # Return the timestamp
    return _datetime.timestamp()


def str_from_ts(ts: float, as_local: bool = False) -> str:
    """
    Convert from a timestamp (seconds since epoch) to a human-readable string (ISO 8601 without a time zone).

    Args:
        ts: timestamp
        as_local (optional): If true return local time, default UTC time

    Returns:
        Timestamp as human readable string
    """
    if as_local:
        # Convert to local time (timezone-aware)
        _datetime = dt.datetime.fromtimestamp(ts).astimezone()
    else:
        # Convert to UTC (timezone-aware)
        _datetime = dt.datetime.fromtimestamp(ts, tz=dt.timezone.utc)

    # Format the datetime as a string
    return _datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")


def ts_from_index(idx: int, sample_rate: float, ts_offset_sec: int = 0) -> float:
    """
    Convert from sample idx to timestamp

    Args:
        idx: Sample index (first sample is index 0)
        sample_rate: Samples per seconds
        ts_offset_sec (optional): Timestamp in seconds since Epoch (1970:01:10T00:00:00)
            ts_offset is the timestamp corresponding to index 0,
            by default this is 0, implying that indexing starts at Epoch (1970:01:10T00:00:00)

    Returns
        Timestamp corresponding to given sample index

    """
    return (idx / sample_rate) + ts_offset_sec


def index_from_ts(ts: float, sample_rate: float, ts_offset_sec: int = 0) -> int:
    """
    Convert from timestamp to sample index

    Args:
    ts: timestamp in seconds from Epoch (1970:01:10T00:00:00)
    sample_rate: samples per seconds
    ts_offset_sec (optional): timestamp in seconds since Epoch (1970:01:10T00:00:00)
        ts_offset is the timestamp corresponding to index 0,
        by default this is 0, implying that indexing starts at Epoch (1970:01:10T00:00:00)

    Returns:
        sample index (first sample is index 0)
    """

    return int((ts - ts_offset_sec) * sample_rate)



"""A collection of functions and information for the NOSTRA system."""

import numpy as np

from pyant.models import FiniteCylindricalParabola, FiniteCylindricalParabolaParams


def generate_nostra() -> tuple[FiniteCylindricalParabola, FiniteCylindricalParabolaParams]:
    """Nostra Finite Cylindrical Parabola beam, note that these values might not be correct"""
    beam = FiniteCylindricalParabola()
    params = FiniteCylindricalParabolaParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=224e6,
        width=120.0,
        height=40.0,
        aperture_width=120.0,  # TODO: Should be optional, can be removed once pyant is updated
    )

    return beam, params



"""MU radar object"""

from pyant.beam import Beam
from pyant.types import Parameters
from radardef.components.radar_station_template import RadarStation
from radardef.radar_stations.pansy.beams import (
    pansy_array_beam,
    pansy_interpolated_array_beam,
)
from radardef.types import StationID


class Pansy(RadarStation):
    """
    Pansy radar definition

    Args:
        interpolated (optional): If beam should be interpolated, by default False.

    """

    def __init__(self, interpolated: bool = False) -> None:

        beam: Beam
        params: Parameters
        if interpolated:
            beam, params = pansy_interpolated_array_beam()
        else:
            beam, params = pansy_array_beam()

        super().__init__(
            StationID.PANSY,
            transmitter=True,
            receiver=True,
            lat=-69.0066066316,
            lon=39.5930902267,
            alt=0.0,
            beam=beam,
            beam_parameters=params,
            power=0.5e6,
            power_per_element=500.0,
            frequency=47e6,
        )



"""Eiscat VHF radar object"""

from radardef.components.radar_station_template import RadarStation
from radardef.radar_stations.eiscat.beams.vhf import eiscat_vhf_beam
from radardef.types import StationID

from .converters import EiscatMatbzToDrf
from .data_loaders import DrfLoader
from .validators import EiscatMatlab


class EiscatVHF(RadarStation):
    """Eiscat VHF radar definition"""

    def __init__(self) -> None:
        beam, params = eiscat_vhf_beam()
        super().__init__(
            StationID.EISCAT_VHF,
            transmitter=True,
            receiver=True,
            lat=69.5866115,
            lon=19.221555,
            alt=85.0,
            beam=beam,
            beam_parameters=params,
            noise_temperature=100,
            power=1.6e6,
            frequency=224e6,
            converters=[EiscatMatbzToDrf()],
            validator=EiscatMatlab(),
            data_loaders=[DrfLoader()],
        )



"""
DataLoaderCollection gathers all given radars and extracts their data loaders, the collection can then
load data from any file given a compatible data loader is available.
"""

import copy
import logging
from pathlib import Path

import numpy as np

from radardef.components.data_loader_template import DataLoader
from radardef.components.radar_station_template import RadarStation
from radardef.types import TargetFormat


class DataLoaderCollection:
    """
    Collection of all available data loaders from the available radar objects,
    provides a way to gather and provide loaders for many different formats in one place.

    Args:
        radars: List of radars that the collections data loaders will be extracted from.

    """

    __logger = logging.getLogger(__name__)

    def __init__(self, radars: list[RadarStation]):
        self.__data_loaders: dict[TargetFormat, DataLoader] = dict()

        for radar in radars:
            for data_loader in radar.get_data_loaders():
                self._register_data_loader(data_loader)

    def load_data(
        self,
        path: Path,
        converted_format: TargetFormat = TargetFormat.UNKNOWN,
    ) -> DataLoader | None:
        """
        Get a dataloader compatible with the file at the path.

        Args:
            path: Path to file that should be loaded
            converted_format (optional): Format of the converted data,
                if not specified the format will be found by the validator.

        Returns:
            A compatible dataloader if available, otherwise None

        """

        if converted_format is TargetFormat.UNKNOWN:
            converted_format = self._get_load_format(path)
        try:
            loader = copy.deepcopy(self.__data_loaders[converted_format])
            loader.load(path)
            return loader
        except KeyError:
            self.__logger.info(
                f"No dataloader available for format: {converted_format}, files affected: {path}"
            )
        return None

    def _get_load_format(self, path: Path) -> TargetFormat:
        """Determine converted format of given file."""

        for converted_format, data_loader in self.__data_loaders.items():
            if data_loader.validate(path):
                return converted_format

        self.__logger.warning("Could not find any matching validator for the given format")
        return TargetFormat.UNKNOWN

    def _register_data_loader(self, data_loader: DataLoader) -> None:
        """Register a data loader to the collection."""
        self.__data_loaders[data_loader.converted_format] = data_loader



"""
Main entry for the CLI functionality
"""

import argparse
import logging
import sys
from collections.abc import Callable

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
except ImportError:

    class COMM_WORLD:
        rank = 0
        size = 1

    comm = COMM_WORLD()  # type: ignore[assignment]

from ..tools import profiling

logger = logging.getLogger(__name__)

COMMANDS: dict[str, dict] = dict()


def build_parser() -> argparse.ArgumentParser:
    """Build default CLI parser including arguments and subparsers."""

    parser = argparse.ArgumentParser(description="Radar data tool box")

    parser.add_argument("-p", "--profiler", action="store_true", help="Run profiler")
    parser.add_argument("-v", "--verbose", help="Increase output verbosity", action="count", default=0)

    subparsers = parser.add_subparsers(help="Available command line interfaces", dest="command")
    subparsers.required = True

    # Adds any item in the COMMAND dict as a subparser
    for name, dat in COMMANDS.items():
        parser_builder, command_help = dat["parser"]
        cmd_parser = subparsers.add_parser(name, help=command_help)
        parser_builder(cmd_parser)

    return parser


def add_command(
    name: str,
    function: Callable,
    parser_build: Callable[[argparse.ArgumentParser], argparse.ArgumentParser],
    command_help: str = "",
) -> None:
    """
    Adds a parser with related function to the global COMMAND dict.

    Args:
        name: Command name
        function: Main function of the callable argument
        parser_build: Parser builder, adds needed commands
        command_help: Command help text

    """

    global COMMANDS
    COMMANDS[name] = dict()
    COMMANDS[name]["function"] = function
    COMMANDS[name]["parser"] = (parser_build, command_help)


def main() -> None:
    """Create and execute CLI interface."""

    parser = build_parser()
    args = parser.parse_args()

    handler = logging.StreamHandler(sys.stdout)
    if comm.size > 1:
        formatter = logging.Formatter(f"RANK{comm.rank} - %(asctime)s %(levelname)s %(name)s - %(message)s")
    else:
        formatter = logging.Formatter("%(asctime)s %(levelname)s %(name)s - %(message)s")
    handler.setFormatter(formatter)

    lib_logger = logging.getLogger("radardef")
    if args.verbose > 0:
        lib_logger.addHandler(handler)
        lib_logger.setLevel(logging.INFO)
    else:
        logger.addHandler(handler)
        logger.setLevel(logging.WARNING)

    if args.verbose > 1:
        logger.info("Logging level set to debug")
        lib_logger.setLevel(logging.DEBUG)

    if args.profiler:
        profiling.profile()

    function = COMMANDS[args.command]["function"]
    logger.info(f"Executing command {args.command}")

    function(args, logger)

    if args.verbose > 0 and args.profiler:
        stats, total = profiling.get_profile()
        profiling.print_profile(stats, total=total)
        profiling.profile_stop()



import logging
from pathlib import Path

from radardef.collections import (
    ConverterCollection,
    DataLoaderCollection,
    FormatCollection,
)
from radardef.components import DataLoader, RadarStation
from radardef.radar_stations import ESR, TSDR, Eiscat3D, EiscatUHF, EiscatVHF, Mu, Pansy
from radardef.types import SourceFormat, TargetFormat
from radardef.types import Eiscat3DLocation, EiscatUHFLocation, DishDiameter


class RadarDef:
    """
    A tool to expose the radar station data processing components from several radar stations in one place.
    Loading and converting data from multiple formats at once, radar specifications from all radar stations
    available in one place.
    """

    __logger = logging.getLogger(__name__)

    @property
    def radar_stations(self) -> list[str]:
        """All available radarstations"""
        return list(self.__radars.keys())

    @property
    def converter_collection(self) -> ConverterCollection:
        """Collection of all converters"""
        return self.__converter_collection

    @property
    def data_loader_collection(self) -> DataLoaderCollection:
        """Collection of all data loader"""
        return self.__data_loader

    @property
    def format_collection(self) -> FormatCollection:
        """Collection for format validation"""
        return self.__format_collection

    def __init__(self) -> None:

        self.__radars: dict[str, RadarStation] = dict()

        for location_3d in Eiscat3DLocation:
            self.add_radar(Eiscat3D(location_3d))
        for location_uhf in EiscatUHFLocation:
            self.add_radar(EiscatUHF(location_uhf))
        for dish_diameter in DishDiameter:
            self.add_radar(ESR(dish_diameter))

        self.add_radars([Mu(), Pansy(), TSDR(), EiscatVHF()])

        self.reload_collections()

    def reload_collections(self) -> None:
        """Reload all collections, if new radar stations has been added"""
        self.__converter_collection = ConverterCollection(list(self.__radars.values()))
        self.__format_collection = FormatCollection(list(self.__radars.values()))
        self.__data_loader = DataLoaderCollection(list(self.__radars.values()))

    def add_radar(self, radar_station: RadarStation) -> None:
        """Add one radar station to the collection, then reload the collections"""
        self.__radars[radar_station.station_id.lower()] = radar_station
        self.reload_collections()

    def add_radars(self, radar_stations: list[RadarStation]) -> None:
        """Add several radar stations to the collection, then reload the collections"""
        for radar in radar_stations:
            self.__radars[radar.station_id.lower()] = radar
        self.reload_collections()

    def delete_radar(self, key: str) -> None:
        """Remove a radar station, key needs to match station id"""

        try:
            del self.__radars[key]
        except KeyError:
            self.__logger.info("No radar deleted, key does not exist")
        self.reload_collections()

    def get_radar(self, id: str) -> RadarStation | None:
        """Get a specific radar station, key needs to match station id"""
        try:
            return self.__radars[id.lower()]
        except KeyError:
            self.__logger.info(f"No radar available with id: {id}")
            return None

    def get_source_format(self, path: Path) -> SourceFormat:
        """Source format of the given path"""
        return self.__format_collection.get_format(path)

    def is_source_format(self, path: Path, source_format: SourceFormat) -> bool:
        """Determines if the path is the given source format"""
        return self.__format_collection.is_format(path, self._validate_source_format(source_format))

    def convert(
        self,
        raw_paths: list[str] | list[Path] | str | Path,
        target_format: TargetFormat,
        output_directory: str,
    ) -> list[Path] | None:
        """
        Convert data to a target format

        Args:
            raw_paths: One or several paths to files/folders to convert.
                Eiscat matlab files must be a folder and not a specific file
            target_format: Format to convert to
            output_directory: Path to output directory to store the converted data

        Returns:
            Paths to the converted files

        """

        if not isinstance(raw_paths, list):
            raw_paths = [raw_paths]  # type: ignore[assignment]

        roots = self._get_root_directories(raw_paths)  # type: ignore[arg-type]
        path_and_format = self._get_source_formats(roots, self.__format_collection)
        try:
            paths, source_formats = zip(*path_and_format)
        except ValueError:
            self.__logger.error("Input path/paths is not a valid file/directory")
            return None

        return self.__converter_collection.convert(
            paths,
            source_formats,
            self._validate_target_format(target_format),
            Path(output_directory).resolve(),
        )

    def load_data(
        self,
        path: Path | str,
        converted_format: TargetFormat = TargetFormat.UNKNOWN,
    ) -> DataLoader | None:
        """
        Load data from a converted file

        Args:
            path: Path to file to load data from
            converted_format (optional): Converted format of the data,

        Returns:
            DataLoader
        """

        return self.__data_loader.load_data(
            Path(path).resolve(),
            self._validate_target_format(converted_format),
        )

    def _validate_source_format(self, source_format: SourceFormat) -> SourceFormat:
        """Makes sure that the input is of kind SourceFormat"""
        try:
            source_format = SourceFormat(source_format.lower())
        except ValueError:
            source_format = SourceFormat.UNKNOWN
        return source_format

    def _validate_target_format(self, target_format: TargetFormat) -> TargetFormat:
        """Makes sure that the input is of kind TargetFormat"""
        try:
            target_format = TargetFormat(target_format.lower())
        except ValueError:
            target_format = TargetFormat.UNKNOWN
        return target_format

    def _get_root_directories(self, raw_paths: list[str] | list[Path]) -> list[Path]:
        """
        Returns a list of root directories for any directory given and
        any path to a file

        Args:
            paths: Paths to files and directories

        Returns:
            list containing the files listed in the input path and
            the root directories for each directory path listed
        """

        def root_folders(path: Path) -> list[Path]:
            """Returns the root folders of a directory"""
            root_directories = []
            # if folder is a root folder
            if path.is_dir() and not [i for i in path.iterdir() if i.is_dir()]:
                root_directories.append(path)
            # else find root
            else:
                for f in path.iterdir():
                    if f.is_dir():
                        if not [i for i in f.iterdir() if i.is_dir()]:
                            root_directories.append(f)
                        else:
                            root_directories.extend(root_folders(f))
            return root_directories

        dirs = []
        for raw_path in raw_paths:
            path = Path(raw_path).resolve()
            if path.is_dir():
                dirs.extend(root_folders(path))
            elif path.is_file():
                dirs.append(path)
        return dirs

    def _get_source_formats(
        self, paths: list[Path], format_validator: FormatCollection
    ) -> list[tuple[Path, SourceFormat]]:
        """
        Extract format for a folder (assuming all types in the folder has
        the same format and no subfolders) or individual files

        Args:
            paths: List of paths to directories and files
            format_validator: FormatCollection object that contains format checkers

        """

        source_formats = []
        for path in paths:
            if path.is_dir():
                files = [f for f in path.iterdir() if f.is_file()]
                if files is not None:
                    source_formats.append((path, format_validator.get_format(files[0])))
                else:
                    source_formats.append((path, SourceFormat.UNKNOWN))
            elif path.is_file():
                source_formats.append((path, format_validator.get_format(path)))

        return source_formats



"""A collection of functions and information for the EISCAT VHF Radar system."""

import numpy as np

from pyant.models import FiniteCylindricalParabola, FiniteCylindricalParabolaParams


def eiscat_vhf_beam() -> tuple[FiniteCylindricalParabola, FiniteCylindricalParabolaParams]:
    """
    Eiscat VHF with all panels moving as a whole [1]_.

    NOTE:
        Has an extra method called :code:`calibrate` that numerically calculates
        the integral of the gain and scales the gain pattern according.

    """
    beam = FiniteCylindricalParabola()
    params = FiniteCylindricalParabolaParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=224e6,
        width=120.0,
        height=40.0,
        aperture_width=120.0,  # TODO: Should be optional, can be removed once pyant is updated
    )
    return beam, params



import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.coordinates import AltAz, EarthLocation, Galactic
from astropy.time import Time
from astropy_healpix import HEALPix
from pygdsm import GlobalSkyModel16

from radardef.components.radar_station_template import RadarStation


def temperature_map(
    radar: RadarStation, time: Time, res: int, flatten: bool = False
) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
    """Temperature map"""
    az_mat, el_mat = np.meshgrid(
        np.linspace(0, 2 * np.pi, num=res),
        np.linspace(0, np.pi / 2, num=res),
    )
    shape = az_mat.shape
    az_mat = az_mat.reshape((az_mat.size,))
    el_mat = el_mat.reshape((el_mat.size,))

    gsm = GlobalSkyModel16(freq_unit="Hz", include_cmb=True)
    mm = gsm.generate(radar.frequency)
    hp_obj = HEALPix(nside=gsm.nside, order="RING", frame=Galactic)
    coords = AltAz(
        az=az_mat * u.rad,
        alt=el_mat * u.rad,
        obstime=time,
        location=EarthLocation(
            lat=radar.lat * u.deg,
            lon=radar.lon * u.deg,
        ),
    )
    temp_mat = hp_obj.interpolate_bilinear_skycoord(coords, mm)
    if not flatten:
        az_mat = az_mat.reshape(shape)
        el_mat = el_mat.reshape(shape)
        temp_mat = temp_mat.reshape(shape)
    return az_mat, el_mat, temp_mat


def calculate_antenna_temperature(radar: RadarStation, time: Time, res: int) -> float:
    """Calculate antena temperature"""
    az_mat, el_mat, temp_mat = temperature_map(radar, time, res, flatten=True)
    g_mat = radar.beam.sph_gain(
        azimuth=az_mat,
        elevation=el_mat,
        parameters=radar.beam_parameters,
    )
    g_mat = g_mat.flatten()  # type: ignore
    d_az = 2 * np.pi / res
    d_el = np.pi / 2 / res
    d_S = np.cos(el_mat) * d_az * d_el

    # ugly numerical integration
    g_int = np.sum(g_mat * d_S)
    t_int = np.sum(temp_mat * g_mat * d_S)

    return t_int / g_int



"""Eiscat mat.bz2 validator"""

from pathlib import Path

from radardef.components.validator_template import Validator
from radardef.types import SourceFormat


class EiscatMatlab(Validator):
    """Eiscat mat.bz2 validator"""

    def __init__(self) -> None:
        super().__init__(SourceFormat.EISCAT_MATBZ)

    def validate(self, src: str | Path) -> bool:
        """Validate file is a Eiscat mat.bz2 file"""
        path = Path(src).resolve()
        return path.name.endswith(".mat.bz2")



from .eiscat_3d_beam import (eiscat_3d_single_subarray_beam,
                             eiscat_3d_stage1_beam,
                             eiscat_3d_stage1_interp_beam,
                             eiscat_3d_stage2_beam,
                             eiscat_3d_stage2_interp_beam)



"""Tromso Space Debris Radar (TSDR) object"""

from pyant.beam import Beam
from pyant.types import Parameters
from radardef.components.radar_station_template import RadarStation
from radardef.radar_stations.eiscat.beams.tsdr import tsdr_beam, tsdr_phased_beam
from radardef.types import StationID

from .converters import EiscatMatbzToDrf
from .data_loaders import DrfLoader
from .validators import EiscatMatlab


class TSDR(RadarStation):
    """
    Tromso Space Debris Radar (TSDR) definition

    Args:
        phased: If the beam should be phased
    """

    def __init__(self, phased: bool = False) -> None:
        beam: Beam
        params: Parameters
        if phased:
            beam, params = tsdr_phased_beam()
        else:
            beam, params = tsdr_beam()

        super().__init__(
            StationID.TSDR,
            transmitter=True,
            receiver=True,
            lat=69.5866115,
            lon=19.221555,
            alt=85.0,
            beam=beam,
            beam_parameters=params,
            noise_temperature=100,
            power=0.5e6,
            frequency=1.8e9,
            converters=[EiscatMatbzToDrf()],
            validator=EiscatMatlab(),
            data_loaders=[DrfLoader()],
        )



"""Wrappers to simplify the usage of MPI on individual functions"""

import logging
from functools import wraps
from typing import Any, Callable

logger = logging.getLogger(__name__)

logger.debug("Importing MPI")
try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
except ImportError:
    logger.debug("MPI import failed: reverting to one process")

    class COMM_WORLD:
        rank = 0
        size = 1

    comm = COMM_WORLD()  # type: ignore[assignment]


def MPI_target_arg(arg_index: int, MPI: bool = False, MPI_root: int = 0) -> Any:
    """
    Decorates the target function to parallelize over a single input iterable
    positional argument identified by the given index.


    Args:
        arg_index: index of the argument to iterate over
        MPI (optional): if MPI should be enabled, default is false
        MPI_root (optional): Which process to gather the results in (tasks will be divided on the rest).
            If negative no results will be gathered.
    """

    def _mpi_wrapper(func: Callable) -> Any:
        """Wrapper"""

        @wraps(func)
        def _mpi_wrapped_func(*args: Any, **kwargs: Any) -> Any:
            """Parallelizes the function over the iterable arg."""
            input_list = args[arg_index]
            _args = list(args)
            rets = [None] * len(input_list)

            if MPI:
                iter_inds = range(comm.rank, len(input_list), comm.size)
            else:
                iter_inds = range(len(input_list))

            for ind in iter_inds:
                _args[arg_index] = input_list[ind]
                rets[ind] = func(*_args, **kwargs)

            if MPI and MPI_root is None:
                pass
            elif MPI and MPI_root >= 0:
                if comm.rank == MPI_root:
                    for thr_id in range(comm.size):
                        if thr_id == MPI_root:
                            continue

                        mpi_inds = range(thr_id, len(input_list), comm.size)
                        for ind in mpi_inds:
                            rets[ind] = comm.recv(source=thr_id, tag=ind)
                else:
                    for ind in iter_inds:
                        comm.send(rets[ind], dest=MPI_root, tag=ind)

                if comm.rank != MPI_root:
                    for ind in iter_inds:
                        rets[ind] = None
            elif MPI and MPI_root < 0:
                for ind in iter_inds:
                    rets[ind] = comm.bcast(rets[ind], root=comm.rank)

            return rets

        return _mpi_wrapped_func

    return _mpi_wrapper


def MPI_target_args(
    arg_indexes: range, loading_bar: bool = False, MPI: bool = False, MPI_root: int = 0
) -> Any:
    """
    Decorates the target function to parallelize over several iterable
    positional arguments identified by the index range.


    Args:
        arg_indexes: The indexes to iterate over, note that the length of the arguments to
            iterate over has to be the same size.
        MPI (optional): If MPI should be enabled, default is false
        MPI_root (optional): Which process to gather the results in (tasks will be divided on the rest).
            If negative no results will be gathered.
    """

    def _mpi_wrapper(func: Callable) -> Any:
        """Wrapper"""

        @wraps(func)
        def _mpi_wrapped_func(*args: Any, **kwargs: int) -> list[Any]:
            """Parallelizes the function over the iterable args."""
            iterable_arguments = [args[arg] for arg in arg_indexes]
            try:
                number_of_indexes = len(iterable_arguments[0])
            except TypeError:
                return func(*args, **kwargs)

            # Validate that the arguments are of the same length
            for argument_list in iterable_arguments:
                if len(argument_list) is not number_of_indexes:
                    raise ValueError("The arguments length are not of the same size")

            func_calls = [None] * number_of_indexes
            func_rets = [None] * number_of_indexes
            tmp_args = list(args)
            for i in range(number_of_indexes):
                for arg in arg_indexes:
                    tmp_args[arg] = iterable_arguments[arg - min(arg_indexes)][i]
                func_calls[i] = func(*tmp_args, **kwargs)

            if MPI:
                iter_inds = range(comm.rank, number_of_indexes, comm.size)
            else:
                iter_inds = range(number_of_indexes)

            if MPI and MPI_root >= 0:
                if comm.rank == MPI_root:
                    for thr_id in range(comm.size):
                        if thr_id == MPI_root:
                            continue

                        mpi_inds = range(thr_id, number_of_indexes, comm.size)
                        for ind in mpi_inds:
                            func_rets[ind] = comm.recv(source=thr_id, tag=ind)
                else:
                    for ind in iter_inds:
                        comm.send(func_calls[ind], dest=MPI_root, tag=ind)

                if comm.rank != MPI_root:
                    for ind in iter_inds:
                        func_rets[ind] = None
            elif MPI and MPI_root < 0:
                for ind in iter_inds:
                    comm.bcast(func_calls[ind], root=comm.rank)
            elif not MPI:
                for ind in iter_inds:
                    func_rets[ind] = func_calls[ind]
            return func_rets

        return _mpi_wrapped_func

    return _mpi_wrapper



"""
The CLI download functionality, abstracts the download functionality to a user friendly CLI interface
"""

import argparse
import logging

from radardef.download.eiscat.download import download

from .commands_cli import add_command


def main(args: argparse.Namespace, cli_logger: logging.Logger) -> None:
    """Download CLI available to the user"""
    download(
        args.day,
        args.mode,
        args.instrument,
        args.dst,
        logger=cli_logger,
        progress=args.progress,
        wget=args.wget,
    )


def parser_build(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Adds mandatory and optional positional arguments to the parser."""

    parser.add_argument("day", help='Experiment day, e.g. "20210412"')
    parser.add_argument("mode", help='Experiment mode, e.g. "leo_bpark_2.1u_NO"')
    parser.add_argument("instrument", help='Experiment instrument, e.g. "uhf|32m|42m"')
    parser.add_argument("dst", default=".", help="destination folder")
    parser.add_argument(
        "-p",
        "--progress",
        action="store_true",
        help="show progressbar for download",
    )
    parser.add_argument(
        "--wget",
        action="store_true",
        help="use wget instead of requests for download",
    )

    return parser


add_command(
    name="download",
    function=main,
    parser_build=parser_build,
    command_help="Download data files from the given radar (Only Eiscat is supported for now)",
)



"""A collection of functions and information for the EISCAT 3D Radar system.

Notes
-----
Configuration are taken from [1]_.


.. [1] (Technical report) Vierinen, J., Kastinen, D., Kero, J.,
    Grydeland, T., McKay, D., Roynestad, E., Hesselbach, S., Kebschull, C., &
    Krag, H. (2019). EISCAT 3D Performance Analysis

"""

import pathlib
from pathlib import Path
from typing import Any, Optional

import numpy as np
import numpy.typing as npt
import scipy.constants

from pyant.models import Array, ArrayParams, InterpolatedArray, InterpolatedArrayParams
from radardef.tools.types import NDArray_2xN, NDArray_N

from .data import DATA_PATHS

e3d_frequency = 233e6
e3d_antenna_gain = 10.0**0.3  # 3 dB peak antenna gain?


def e3d_subarray(frequency: float) -> tuple[NDArray_N, NDArray_N, NDArray_N]:
    """Generate cartesian positions `x,y,z` in meters of antenna elements in
    one standard EISCAT 3D subarray.
    """
    l0 = scipy.constants.c / frequency

    dx = 1.0 / np.sqrt(3)
    dy = 0.5

    xall = []
    yall = []

    x0_p1 = np.arange(-2.5, -5.5, -0.5).tolist()
    x0_p2 = np.arange(-4.5, -2.0, 0.5).tolist()
    x0 = np.array([x0_p1 + x0_p2])[0] * dx
    y0 = np.arange(-5, 6, 1) * dy

    for iy in range(11):
        nx = 11 - np.abs(iy - 5)
        x_now = x0[iy] + np.array(range(nx)) * dx
        y_now = y0[iy] + np.array([0.0] * (nx))
        xall += x_now.tolist()
        yall += y_now.tolist()

    x = l0 * np.array(xall)
    y = l0 * np.array(yall)
    z = x * 0.0

    return x, y, z


# TODO:Better type hint
def e3d_array(frequency: float, fname: Optional[str] = None, configuration: str = "full") -> npt.NDArray:
    """Generate the antenna positions for a EISCAT 3D Site based on submodule
    positions of a file.
    """

    def _read_e3d_submodule_pos(string_data: str) -> NDArray_2xN:
        """read e3d submodule position and create a 2xN numpy float array"""
        dat = []
        file = string_data.split("\n")
        for line in file:
            if len(line) == 0:
                continue
            dat.append(list(map(lambda x: float(x), line.split())))
        return np.array(dat)

    assert "e3d_subgroup_positions.txt" in DATA_PATHS, "data file missing!"
    path = fname if fname is not None else DATA_PATHS["e3d_subgroup_positions.txt"]

    with open(path, "r") as stream:
        _ant_data = stream.read()

    dat = _read_e3d_submodule_pos(_ant_data)

    sx, sy, sz = e3d_subarray(frequency)

    if configuration == "full":
        pass
    elif configuration == "dense":
        dat = dat[(np.sum(dat**2.0, axis=1) < 27.0**2.0), :]
    elif configuration == "sparse":
        dat = dat[
            np.logical_or(
                np.logical_or(
                    np.logical_and(np.sum(dat**2, axis=1) < 10**2, np.sum(dat**2, axis=1) > 7**2),
                    np.logical_and(np.sum(dat**2, axis=1) < 22**2, np.sum(dat**2, axis=1) > 17**2),
                ),
                np.logical_and(
                    np.sum(dat**2, axis=1) < 36**2,
                    np.sum(dat**2, axis=1) > 30**2,
                ),
            ),
            :,
        ]
    elif configuration == "module":
        dat = np.zeros((1, 2))

    antennas = np.zeros((3, len(sx), dat.shape[0]), dtype=dat.dtype)
    for i in range(dat.shape[0]):
        for j in range(len(sx)):
            antennas[0, j, i] = sx[j] + dat[i, 0]
            antennas[1, j, i] = sy[j] + dat[i, 1]
            antennas[2, j, i] = sz[j]
    return antennas


def eiscat_3d_single_subarray_beam() -> tuple[Array, ArrayParams]:
    """EISCAT 3D Gain pattern for single antenna sub-array."""
    beam = Array(
        antennas=e3d_array(
            e3d_frequency,
            configuration="module",
        ),
    )
    params = ArrayParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=e3d_frequency,
        polarization=beam.polarization.copy(),
    )
    return beam, params


def eiscat_3d_stage1_beam(configuration: str = "dense") -> tuple[Array, ArrayParams]:
    """EISCAT 3D Gain pattern for a dense core of active sub-arrays,
    i.e stage 1 of development.

    Args:
        configuration (optional): Chooses how the stage1 antennas are distributed in the full array,
            alt {'dense', 'sparse'}

    """
    beam = Array(
        antennas=e3d_array(
            e3d_frequency,
            configuration=configuration,
        ),
    )
    params = ArrayParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=e3d_frequency,
        polarization=beam.polarization.copy(),
    )
    return beam, params


# TODO: Clarify Any
def eiscat_3d_stage1_interp_beam(
    path: Optional[str | Path] = None, configuration: str = "dense", **interpolation_kwargs: Any
) -> tuple[InterpolatedArray, InterpolatedArrayParams]:
    """EISCAT 3D Gain pattern for a dense core of active sub-arrays,
    i.e stage 1 of development.

    Args:
        configuration (optional): Chooses how the stage1 antennas are distributed in the full array,
            alt {'dense', 'sparse'}
    """
    beam = InterpolatedArray()
    params = InterpolatedArrayParams(pointing=np.array([0, 0, 1], dtype=np.float64))
    if path is None:
        array, arr_params = eiscat_3d_stage1_beam(configuration=configuration)
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        return beam, params

    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    if path.is_file():
        beam.load(path)
    else:
        array, arr_params = eiscat_3d_stage1_beam(configuration=configuration)
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        beam.save(path)
    return beam, params


def eiscat_3d_stage2_beam() -> tuple[Array, ArrayParams]:
    """EISCAT 3D Gain pattern for a full site of active sub-arrays,
    i.e stage 2 of development.

    """
    beam = Array(
        antennas=e3d_array(
            e3d_frequency,
            configuration="full",
        ),
    )
    params = ArrayParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=e3d_frequency,
        polarization=beam.polarization.copy(),
    )
    return beam, params


# TODO: Clarify Any
def eiscat_3d_stage2_interp_beam(
    path: Optional[str | Path] = None, **interpolation_kwargs: Any
) -> tuple[InterpolatedArray, InterpolatedArrayParams]:
    """EISCAT 3D Gain pattern for a full set of active sub-arrays,
    i.e stage 2 of development.
    """
    beam = InterpolatedArray()
    params = InterpolatedArrayParams(pointing=np.array([0, 0, 1], dtype=np.float64))
    if path is None:
        array, arr_params = eiscat_3d_stage2_beam()
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        return beam, params

    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    if path.is_file():
        beam.load(path)
    else:
        array, arr_params = eiscat_3d_stage2_beam()
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        beam.save(path)
    return beam, params



from radardef.collections import (
    ConverterCollection,
    DataLoaderCollection,
    FormatCollection,
)
from radardef.components import Converter, DataLoader, RadarStation, Validator
from radardef.radar_def import RadarDef
from radardef.radar_stations import *
from radardef.types import (
    Boundparam,
    BoundParams,
    Expparam,
    ExpParams,
    Metadata,
    Metaparam,
)
import types
from radardef.download import download

from .version import __version__



"""Enums to simplify the usage of the radar objects"""

from enum import IntEnum, StrEnum, auto


class SourceFormat(StrEnum):
    """Radar dataformats"""

    MUI = "mui"
    EISCAT_MATBZ = "eiscat_matbz"
    UNKNOWN = "unknown"


class TargetFormat(StrEnum):
    """Formats possible to convert to"""

    H5 = "h5"
    DRF = "drf"
    UNKNOWN = "unknown"


class StationID(StrEnum):
    """Radar station IDs"""

    MU = auto()
    PANSY = auto()
    TSDR = auto()
    ESR_32M = auto()
    ESR_42M = auto()
    EISCAT_3D = auto()
    EISCAT_UHF_TROMSO = auto()
    EISCAT_UHF_KIRUNA = auto()
    EISCAT_UHF_SODANKYLA = auto()
    EISCAT_VHF = auto()
    EISCAT_3D_SKIBOTN = auto()
    EISCAT_3D_KARESUVANTO = auto()
    EISCAT_3D_KAISENIEMI = auto()
    UNKNOWN = auto()


class EiscatUHFLocation(StrEnum):
    """Eiscat UHF locations"""

    TROMSO = auto()
    KIRUNA = auto()
    SODANKYLA = auto()


class Eiscat3DLocation(StrEnum):
    """Eiscat 3D locations"""

    SKIBOTN = auto()
    KARESUVANTO = auto()
    KAISENIEMI = auto()


class Stage(IntEnum):
    """Eiscat 3D radar stages"""

    SINGLE = 0
    STAGE_1 = 1
    STAGE_2 = 2


class BeamType(StrEnum):
    """Beam types"""

    MEASURED = "measured"
    CASSEGRAIN = "cassegrain"


class DishDiameter(StrEnum):
    """Dish diameter"""

    ESR_32M = "32m"
    ESR_42M = "42m"


class Metaparam(StrEnum):
    """Metadata parameter keys"""

    EXPERIMENT = auto()
    BOUNDS = auto()


class Expparam(StrEnum):
    """Experiment parameter keys"""

    NAME = auto()
    TX_PULSE_LENGTH = auto()
    T_RX_START_USEC = auto()
    T_RX_END_USEC = auto()
    T_TX_START_USEC = auto()
    T_TX_END_USEC = auto()
    T_CAL_ON_USEC = auto()
    T_CAL_OFF_USEC = auto()
    RADAR_FREQUENCY = auto()
    RX_CHANNELS = auto()
    TX_CHANNEL = auto()
    T_IPP_USEC = auto()
    T_SAMP_USEC = auto()
    SAMPLE_RATE = auto()
    DATE = auto()
    CODE = auto()
    PULSE = auto()


class Boundparam(StrEnum):
    """Bound parameter keys"""

    TS_START_USEC = auto()
    START = auto()
    TS_END_USEC = auto()
    END = auto()



"""
Package-wide execution time profiling system.

Module containing profiling systems using yappi

"""

import logging
import os
import pathlib
import time
from functools import wraps
from typing import Any, Callable, Optional

try:
    import yappi

    YAPPI_AVAILABLE = True
except ImportError:
    YAPPI_AVAILABLE = False


file_logger = logging.getLogger(__name__)

PACKAGE_PATH = str(pathlib.Path(__file__).parent)
START_TIME: float = 0


def _path_to_module(path: str) -> str:
    """Get path to module"""
    if PACKAGE_PATH not in path:
        return ""
    stem = pathlib.Path(path).stem
    path = "pyant" + os.sep + path.replace(PACKAGE_PATH, "").strip(os.sep)
    module = path.split(os.sep)
    module[-1] = stem
    return ".".join(module)


def check_yappi(func: Callable) -> Callable:
    """Wrapper"""

    def checked_func(*args: Any, **kwargs: Any) -> Callable:
        """Raise error if yappi is not available"""
        if YAPPI_AVAILABLE is False:
            raise ImportError("'yappi' not installed, please install to profile")
        return func(*args, **kwargs)

    return checked_func


@check_yappi
def profile() -> None:
    """Start profile"""
    global START_TIME
    START_TIME = time.time()
    yappi.set_clock_type("cpu")
    yappi.start()


@check_yappi
def get_profile(modules: Optional[list[str]] = None) -> tuple[Any, float]:
    """Get available progile for the module"""

    if modules is None:
        modules = ["radardef"]
    stats = yappi.get_func_stats(
        filter_callback=lambda x: any(list(_path_to_module(x.module).startswith(mod) for mod in modules)),
    )
    stats = stats.sort("ttot", "desc")
    total = time.time() - START_TIME
    return stats, total


def print_profile(stats: Any, total: Optional[int] = None) -> None:
    """Print profile data"""

    header = [
        "Name",
        "Module",
        "Calls",
        "Total [s]",
        "Function [s]",
        "Average [s]",
    ]
    column_sizes = [len(title) for title in header]
    formats = [""] * 3 + ["1.4e"] * 3

    if total is not None:
        header += ["Total [%]"]
        column_sizes += [len(header[-1])]
        formats += ["2.3f"]

    if total is None:
        total = 1

    for ind in range(3, len(header)):
        if column_sizes[ind] < 6:
            column_sizes[ind]

    datas = [
        (
            fn.name,
            _path_to_module(fn.module),
            f"{fn.ncall}",
            fn.ttot,
            fn.tsub,
            fn.tavg,
            fn.ttot / total * 100,
        )
        for fn in stats
    ]
    for data in datas:
        for ind in range(3):
            if column_sizes[ind] < len(data[ind]):
                column_sizes[ind] = len(data[ind])

    _str = " | ".join([f"{title:^{size}}" for title, size in zip(header, column_sizes)])
    print(_str)
    print("-" * len(_str))

    for data in datas:
        _str = " | ".join(
            [f"{x:{fmt}}".ljust(size) for x, size, fmt in zip(data[: len(header)], column_sizes, formats)]
        )
        print(_str)


@check_yappi
def profile_stop(clear: bool = True) -> None:
    """Stop profile"""
    yappi.stop()
    if clear:
        global START_TIME
        START_TIME = 0
        yappi.clear_stats()


def log(logger: logging.Logger, level: int) -> Callable:
    """Decorator that logs the function call at the desired level"""

    def log_func(func: Callable) -> Callable:
        """Wrap functions to print when called"""

        @wraps(func)
        def logged_func(*args: Any, **kwargs: Any) -> Callable:
            """logs functions calls"""
            logger.log(level, f'Function "{func.__name__}" called')
            return func(*args, **kwargs)

        return logged_func

    return log_func



"""Template class for all converters to inherit from"""

from abc import abstractmethod
from pathlib import Path

from radardef.types import SourceFormat, TargetFormat


class Converter:
    """
    Converter template, should be inherited by all converters

    Args:
        source_format: The format the converter takes as input
        target_format: The format the converter converts to

    """

    @property
    def source_format(self) -> SourceFormat:
        """Source format of the data compatible with the converter"""
        return self.__source_format

    @property
    def target_format(self) -> TargetFormat:
        """Format the converter converts to"""

        return self.__target_format

    def __init__(self, source_format: SourceFormat, target_format: TargetFormat) -> None:
        self.__source_format = source_format
        self.__target_format = target_format

    def __str__(self) -> str:
        """Converter source to target specification string"""
        return f"converter from {self.__source_format} to {self.__target_format}"

    @abstractmethod
    def convert(self, src: Path, dst: Path) -> list[Path]:
        """
        Abstract method, convert from source format to target format

        Args:
            src: Path to source directory
            dst: Path to destination directory

        Returns:
            output: List of the generated files directories
        """
        pass



"""MUI format validator"""

from pathlib import Path

from radardef.components.validator_template import Validator
from radardef.types import SourceFormat


class MUI(Validator):
    """MUI format validator"""

    def __init__(self) -> None:
        super().__init__(SourceFormat.MUI)

    def validate(self, src: str | Path) -> bool:
        "File is of format MUI"
        path = Path(src).resolve()
        return len(path.name) == 17 and path.name.startswith("MUI")



"""DRF validator"""

from pathlib import Path

from radardef.components.validator_template import Validator
from radardef.types import TargetFormat


class DRF(Validator):
    """Drf validator"""

    def __init__(self) -> None:
        super().__init__(TargetFormat.DRF)

    def validate(self, src: str | Path) -> bool:
        """Validate path is a directory and of drf format"""
        path = Path(src).resolve()
        return path.is_dir() and path.name.endswith("drf")



"""A collection of functions and information for the Eiscat Svalbard Radar (ESR) system."""

import numpy as np

from spacecoords.spherical import sph_to_cart
from pyant.models import Cassegrain, CassegrainParams


def esr_32m_cassegrain_beam() -> tuple[Cassegrain, CassegrainParams]:
    """ESR 32 meter antenna diameter"""
    beam = Cassegrain(
        peak_gain=10 ** (42.5 / 10),
    )
    params = CassegrainParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=500e6,
        outer_radius=16.0,
        inner_radius=1.73,
    )
    return beam, params


def esr_42m_cassegrain_beam() -> tuple[Cassegrain, CassegrainParams]:
    """ESR 42 meter antenna diameter"""
    beam = Cassegrain(
        peak_gain=10 ** (45.0 / 10),  # Linear gain (42.5 dB)
    )
    params = CassegrainParams(
        # azimut=185.5, elevation=82.1  (since 2019)
        pointing=sph_to_cart(np.array([185.5, 82.1, 1], dtype=np.float64), degrees=True),
        frequency=500e6,
        outer_radius=21.0,  # radius main reflector
        inner_radius=3.3,  # radius subreflector (eyeballed from photo)
    )
    return beam, params



"""A collection of functions and information for the Tromso Space Debris Radar (TSDR) system.

Notes:
    Configurations taken from [1]_.

    .. [1] (White paper) McKay, D., Grydeland, T., Vierinen, J.,
        Kastinen, D., Kero, J., Krag, H. (2019) Conversion of the EISCAT VHF
        antenna into the Tromso Space Debris Radar

"""

import numpy as np

from pyant.models import (
    FiniteCylindricalParabola,
    FiniteCylindricalParabolaParams,
    PhasedFiniteCylindricalParabola,
    PhasedFiniteCylindricalParabolaParams,
)


def tsdr_beam() -> tuple[FiniteCylindricalParabola, FiniteCylindricalParabolaParams]:
    """
    Tromso Space Debris Radar system with all panels moving as a whole [1]_.

    NOTE:
        Has an extra method called :code:`calibrate` that numerically calculates
        the integral of the gain and scales the gain pattern according.



    """
    beam = FiniteCylindricalParabola()
    params = FiniteCylindricalParabolaParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=1.8e9,
        width=120.0,
        height=40.0,
        aperture_width=120.0,  # TODO: Should be optional, can be removed once pyant is updated
    )
    return beam, params


def tsdr_phased_beam() -> tuple[PhasedFiniteCylindricalParabola, PhasedFiniteCylindricalParabolaParams]:
    """
    Tromso Space Debris Radar system with panels moving independently.

    NOTE:
        This model is a list of the 4 panels. This applies heave approximations on
        the behavior of the gain pattern as the panels move. Considering a linefeed
        of a single panel, it will receive more reflection area if one of the
        adjacent panels move in into the same pointing direction therefor
        distorting the side-lobe as support structures pass but also narrowing the
        resulting beam.None of these effects are considered here and this
        approximation is reasonably valid when all panels are pointing in
        sufficiently different directions.

    """
    beam = PhasedFiniteCylindricalParabola()
    params = PhasedFiniteCylindricalParabolaParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        phase_steering=0.0,
        depth=18.0,
        frequency=1.8e9,
        width=120.0,
        height=40.0,
        aperture_width=120.0,  # TODO: Should be optional so can be removed when pyant is updated
    )
    return beam, params



import logging
import os
import shutil
import subprocess
import zipfile
from itertools import chain
from pathlib import Path
from typing import Any, Optional

import requests
from lxml import html
from tqdm import tqdm


def format_bytes(size: float) -> str:
    """Convert bytes to a human-readable format."""
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024.0:
            return f"{size:.2f} {unit}"
        size /= 1024.0

    return ""


# Start
# DAY=20161021 mode=leo_bpark_2.5u_3P
# url
# https://portal.eiscat.se/schedule/tape2.cgi?exp=leo_bpark_2.5u_3P&date=20161021
# url to download page
# https://portal.eiscat.se/schedule/tape2.cgi?date=20161021&exp=leo_bpark_2.5u_3P&dl=1
# dl == download = true
# location-based download tar or zip format
# https://portal.eiscat.se/schedule/download.cgi?r=249696&r=249698&maxtar=4&filename=leo_bpark_2.5u_3P20161021&submit=Location-based&format=tar
# https://portal.eiscat.se/schedule/download.cgi?r=249696&r=249698&maxtar=4&filename=leo_bpark_2.5u_3P20161021&submit=Location-based&format=zip
# zip download link -- all
# https://rebus.eiscat.se:37009/249696;249698/leo_bpark_2.5u_3P20161021.zip


QUERY_URL = "https://portal.eiscat.se/schedule/tape2.cgi"
DOWNLOAD_URL = "https://rebus.eiscat.se:37009"


def get_download_nodes(
    day: str, mode: str, logger: Optional[logging.Logger] = None
) -> list[list[dict[str, Any]]] | None:
    """Extract node identifiers from download page"""
    url = f"{QUERY_URL}?exp={mode}&date={day}&dl=1"
    response = requests.get(url)
    if response.status_code == 200:
        html_content = response.text
        # parse
        tree = html.fromstring(html_content)
        checkboxes = tree.xpath('//input[@type="checkbox"]')
        trs = [cb.xpath("./ancestor::tr[1]") for cb in checkboxes]
        flattened_trs = list(chain.from_iterable(trs))
        product_nodes = []
        nodes = []
        for tr in flattened_trs:
            tds = tr.xpath(".//td")
            node = {}
            # tds[0]
            cb = tds[0].find('.//input[@name="r"]')
            node["id"] = cb.get("value")
            # tds[1]
            node["type"] = type = tds[1].text
            # tds[2]
            node["start"] = tds[2].text
            if type == "data":
                # tds[3]
                node["end"] = tds[3].text
                # tds[5]
                # NOTE: assuming number is given in KB
                node["bytes"] = int(tds[5].text) * 1024
            elif type == "info":
                # tds[4]
                tokens = tds[4].text.split(" ")
                node["exp"] = {"instrument": tokens[1], "mode": tokens[2]}
            nodes.append(node)
            if type == "info":
                product_nodes.append(nodes)
                nodes = []
        return product_nodes
    else:
        if logger:
            logger.error(f"Failed to fetch the HTML content. Status code: {response.status_code}")
        return None


def download(
    day: str,
    mode: str,
    instrument: str,
    dst: str,
    logger: Optional[logging.Logger] = None,
    progress: bool = False,
    wget: bool = False,
) -> str | None:
    """Download zip file with eiscat measurements.

    Example:

    ```python
        path = download('20220408','leo_bpark_2.1u_NO', 'uhf', '/data')
    ```

    This function uses a specific download link, for which data access is
    geo-restricted to IP-addresses in Scandinavia. The link also uses a
    non-standard port number, so it may be blocked by an agressive firewall. For
    data access outside Scandinavia, see https://eiscat.se/

    The triplet ('day', 'mode', 'instrument') must correspond to an actual
    dataset in the Eiscat archive. Search for available datasets is possible
    using the Web portal https://portal.eiscat.se/. If a download is interrupted
    by the user (ctrl-c), the incomplete zip download file will be removed.

    Args:
        day: Day of experiment ("YYYYMMDD")
        mode: Mode identifier
        instrument: Instrument identifier
        dst: Path to folder where downloaded zip file will be stored
        logger (optional): Optional logger object
        progress (optional): Print download progress bar to stdout
        wget (optional): Use wget instead of requests for download, default = False

    Returns:
        The absolute path to the downloaded zip file or None.

    Raises:
        FileNotFoundError: 'dst' does not exist
        ExceptioN:  target product does not exist

    """
    path = Path(dst)
    if not path.is_dir():
        raise FileNotFoundError(f"<dst> '{path}' is not a directory")

    # zip
    zip_filename = f"{mode}-{day}-{instrument}.zip"
    zip_download = path / zip_filename

    # check if downloaded zipfile already exists
    if zip_download.exists():
        if logger:
            logger.info(f"Zipfile exists: {zip_download}")
        zip_download = zip_download
        return str(zip_download)

    # Fetch download nodes
    product_nodes = get_download_nodes(day, mode, logger=logger)

    if (product_nodes == None) or (len(product_nodes) == 0):
        raise Exception("No product nodes")

    # select product
    data_nodes = []
    info_nodes = []
    for nodes in product_nodes:
        data_nodes = [node for node in nodes if node["type"] == "data"]
        info_nodes = [node for node in nodes if node["type"] == "info"]

        # check that data matches given instrument
        _instrument = info_nodes[0]["exp"]["instrument"]
        if instrument.casefold() != _instrument.casefold():
            # mismatch instrument
            continue
        else:
            # found first
            break

    if len(data_nodes) == 0:
        raise Exception("No data nodes for product", product_nodes)

    # Size
    bytes = sum([node["bytes"] for node in data_nodes])
    size = format_bytes(bytes)

    # Url
    node_ids = [node["id"] for node in nodes]
    zip_url = f"{DOWNLOAD_URL}/{';'.join(node_ids)}/{mode}{day}.zip"

    # Download
    if logger:
        logger.info(f"Product: {mode} {day} {size}")
        logger.info(f"Url: {zip_url}")

    completed = False
    code = None

    if wget:
        # WGET
        command = ["wget", zip_url, "-P", str(dst), "-O", zip_filename]
        if not progress:
            command.append("-q")
        try:
            result = subprocess.run(command)
            if result.returncode == 0:
                completed = True
            else:
                code = result.returncode
        except KeyboardInterrupt:
            pass

    else:
        # REQUESTS
        response = requests.get(zip_url, stream=True)
        if response.status_code == 200:
            file_size = int(response.headers.get("Content-Length", 0)) or bytes
            if logger:
                logger.info(f"Download: zipfile: {zip_download}")
            # Use tqdm to create a progress bar
            if progress:
                pbar = tqdm(
                    desc="Downloading Eiscat raw data",
                    total=file_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                )

            with open(zip_download, "wb") as file:
                try:
                    for data in response.iter_content(chunk_size=1024):
                        file.write(data)
                        if progress:
                            pbar.update(len(data))
                    completed = True
                    if logger:
                        logger.info("Download: completed")
                except KeyboardInterrupt:
                    pass
                if progress:
                    pbar.close()
        else:
            code = response.status_code

    if not completed:
        # cleanup
        os.remove(zip_download)
        if logger:
            if code is None:
                logger.warning("Download: terminiated")
            else:
                logger.info(f"Download: fail - status code: {code}")
        return None
    else:
        return str(zip_download)


def extract_zip(
    src: str, dst: str, cleanup: bool = True, logger: Optional[logging.Logger] = None
) -> tuple[bool, dict[str, list[str]] | str]:

    src_path = Path(src)
    dst_path = Path(dst)

    if not src_path.is_file():
        return False, f"Zipfile: missing {src_path}"

    if not dst_path.is_dir():
        return False, f"No dst directory {dst_path}"

    # make temporary extraction folder
    # extract_dir = dst / str(uuid.uuid4())
    # extract_dir.mkdir(exist_ok=True, parents=True)

    if logger:
        logger.info(f"Unzip: start {dst_path}")

    # Extract zip file to random location in dst directory
    with zipfile.ZipFile(src_path, "r") as zip_ref:
        paths = [Path(p) for p in zip_ref.namelist()]
        rootpaths = list(set([p.parts[0] for p in paths]))
        extract_dirs = [dst_path / p for p in rootpaths]
        # could check if extracted dirs already exist
        zip_ref.extractall(dst_path)

    if cleanup:
        os.remove(str(src_path))

    if logger:
        logger.info(f"Unzip: done {extract_dirs}")

    return True, {"paths": [str(p) for p in extract_dirs]}


def move_tree(
    src: str, dst: str, update: bool = False, logger: Optional[logging.Logger] = None
) -> tuple[bool, dict[str, str] | str]:
    src_path = Path(src)
    dst_path = Path(dst)
    if not dst_path.is_dir():
        return False, f"No <dst> directory {dst_path}"

    # move src to dst
    if update and dst_path.is_dir():
        # update
        # move subfolders of src individually - if not exists
        for src_subdir in src_path.iterdir():
            dst_subdir = dst_path / src_subdir.name
            if not dst_subdir.exists():
                shutil.move(str(src_subdir), str(dst_subdir))
    else:
        # move
        shutil.move(str(src_path), str(dst_path))

    return True, {"path": str(dst_path / src_path.name)}


"""
    if __name__ == '__main__':

    day = '20220408'
    mode = 'leo_bpark_2.1u_NO'
    instrument = 'uhf'
    # logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    def test_download_nodes():
        nodes = get_download_nodes(day, mode)
        import pprint
        pprint.pprint(nodes)

    def test_download():
        day = '20101125'
        mode = 'leo_sat_1.0l_EI'
        instrument = '42m'
        dst = "."
        wget = True
        ok, result = download(day, mode, instrument, dst,
      logger=logger, progress=True, wget=wget)
       print(ok, result)

       def test_extract():
       zip = "leo_bpark_2.1u_NO-20220408-uhf.zip"
       dst = "."
       ok, result = extract_zip(zip, dst, logger=logger)
       print(ok, result)

       def test_move():
       dst = "3a79f529-6a94-418d-b42e-cbc676036489"
       src = "dst"
       ok, result = move_tree(src, dst, logger=logger)
       print(ok, result)

       test_download()
       # test_extract()
        # test_move()
"""
